import copy
from typing import Iterable, Iterator, List, NoReturn, Optional, Type, Union

from . import abc
from .utilities.inspect import qualified_name
from .utilities.types import Null


class Types(abc.Types):
    def __init__(
        self,
        items: Optional[
            Union[
                abc.Types,
                Iterable[Union[type, abc.Property]],
                type,
                abc.Property,
            ]
        ] = None,
    ) -> None:
        if isinstance(items, (type, abc.Property)):
            items = (items,)
        self._list: List[Union[type, abc.Property]] = []
        if items is not None:
            self._extend(items)

    @classmethod
    def _raise_immutable_type_error(cls) -> NoReturn:
        raise TypeError(
            f"Instances of `{qualified_name(cls)}` cannot be " "modified."
        )

    def __copy__(self) -> abc.Types:
        return self.__class__(self)

    def __deepcopy__(self, memo: Optional[dict] = None) -> abc.Types:
        value: Union[type, abc.Property]
        return self.__class__(
            tuple(
                getattr(copy, "deepcopy")(value, memo=memo) for value in self
            )
        )

    def __repr__(self) -> str:
        # Represent the class by it's fully-qualified type name
        representation = [qualified_name(type(self)) + "("]
        # If it is not empty--we represent the values as a `list`
        if self:
            representation[0] += "["
            for value in self.__iter__():
                value_representation: str = "\n".join(
                    "    " + line
                    for line in (
                        qualified_name(value)
                        if isinstance(value, type)
                        else repr(value)
                    ).split("\n")
                )
                representation.append(f"{value_representation},")
            representation[-1] = representation[-1].rstrip(",")
            representation.append("]")
        representation[-1] += ")"
        # Return a single-line representation if this is an empty list of
        # types, otherwise--break it into multiple lines
        return (
            "\n".join(representation)
            if len(representation) > 2
            else "".join(representation)
        )

    def __contains__(
        self, item: Union[type, abc.Property]  # type: ignore
    ) -> bool:
        return self._list.__contains__(item)

    def __iter__(  # type: ignore
        self,
    ) -> Iterator[Union[type, abc.Property]]:
        return self._list.__iter__()

    def __len__(self) -> int:
        return self._list.__len__()

    def _append(self, value: Union[type, abc.Property]) -> None:
        _validate_type_or_property(value)
        self._list.append(value)

    def _extend(
        self,
        values: Union[Iterable[Union[type, abc.Property]], abc.Types],
    ) -> None:
        value: Union[type, abc.Property]
        for value in values:
            self._append(value)

    def __getitem__(self, index: int) -> Union[type, abc.Property]:
        return self._list.__getitem__(index)


class MutableTypes(Types, abc.MutableTypes):
    """
    Instances of this class are lists which will only take values which are
    valid types for describing a property definition.

    Parameters:

    - items ([type|sob.properties.Property])
    - mutable (bool)
    """

    def __setitem__(
        self, index: int, value: Union[type, abc.Property]
    ) -> None:
        _validate_type_or_property(value)
        self._list.__setitem__(index, value)

    def append(self, value: Union[type, abc.Property]) -> None:
        self._append(value)

    def extend(self, values: Iterable[Union[type, abc.Property]]) -> None:
        self._extend(values)

    def __delitem__(self, index: Union[int, slice]) -> None:
        self._list.__delitem__(index)

    def pop(self, index: int = -1) -> Union[type, abc.Property]:
        return self._list.pop(index)

    def __iadd__(
        self, other: Iterable[Union[type, abc.Property]]
    ) -> abc.MutableTypes:
        self.extend(other)
        return self

    def __add__(
        self, other: Iterable[Union[type, abc.Property]]
    ) -> abc.MutableTypes:
        new_instance: abc.Types = copy.copy(self)
        assert isinstance(new_instance, abc.MutableTypes)
        new_instance.extend(other)
        return new_instance


def _validate_type_or_property(
    type_or_property: Union[Type[abc.Property], abc.Property]
) -> None:
    if not (
        isinstance(type_or_property, (type, abc.Property))
        and (
            (type_or_property is Null)
            or (
                isinstance(type_or_property, type)
                and issubclass(type_or_property, abc.MARSHALLABLE_TYPES)
            )
            or isinstance(type_or_property, abc.Property)
        )
    ):
        raise TypeError(type_or_property)



"""
`sob` is an object serialization/deserialization library intended to facilitate
authoring of API models which are readable and introspective, and to expedite
code and data validation and testing.
"""
from typing import List

from . import utilities
from . import errors
from . import abc
from . import model
from . import meta
from . import types
from . import hooks
from . import properties
from . import test
from . import request
from . import thesaurus

__all__: List[str] = [
    "utilities",
    "abc",
    "model",
    "errors",
    "properties",
    "meta",
    "types",
    "hooks",
    "test",
    "request",
    "thesaurus",
]



import collections
import json as _json
from typing import Any, List, Optional, Sequence, Set, Tuple
from warnings import warn

import yaml as yaml_

from . import abc, meta
from .errors import ObjectDiscrepancyError
from .model import serialize, validate
from .utilities import qualified_name
from .utilities.assertion import (
    assert_in,
    assert_is_instance,
)
from .utilities.types import NoneType


def _get_object_property_names(object_: abc.Object) -> Set[str]:
    meta_: Optional[abc.Meta] = meta.read(object_)
    assert isinstance(meta_, (abc.ObjectMeta, NoneType))
    return set(
        meta_.properties.keys()
        if ((meta_ is not None) and (meta_.properties is not None))
        else ()
    )


def _object_discrepancies(
    object_a: abc.Object, object_b: abc.Object
) -> abc.OrderedDict[
    str,
    Tuple[Optional[abc.MarshallableTypes], Optional[abc.MarshallableTypes]],
]:
    discrepancies: abc.OrderedDict[
        str,
        Tuple[
            Optional[abc.MarshallableTypes], Optional[abc.MarshallableTypes]
        ],
    ] = collections.OrderedDict()
    for property_ in sorted(
        (
            _get_object_property_names(object_a)
            | _get_object_property_names(object_b)
        ),
        key=lambda item: item[0],
    ):
        a_value: Optional[abc.MarshallableTypes]
        b_value: Optional[abc.MarshallableTypes]
        try:
            a_value = getattr(object_a, property_)
        except AttributeError:
            a_value = None
        try:
            b_value = getattr(object_b, property_)
        except AttributeError:
            b_value = None
        if a_value != b_value:
            discrepancies[property_] = (a_value, b_value)
    return discrepancies


def _get_value_discrepancies_error_message(
    class_name: str,
    property_name: str,
    property_value_a: Any,
    property_value_b: Any,
) -> str:
    a_serialized = serialize(property_value_a)
    b_serialized = serialize(property_value_b)
    a_representation = repr(property_value_a)
    b_representation = repr(property_value_b)
    a_representation = "".join(
        line.strip() for line in a_representation.split("\n")
    )
    b_representation = "".join(
        line.strip() for line in b_representation.split("\n")
    )
    message: List[str] = []
    if a_serialized != b_serialized or (a_representation != b_representation):
        message.append(
            "\n    %s().%s:\n        %s        \n        %s\n     "
            "   %s"
            % (
                class_name,
                property_name,
                a_serialized,
                "==" if a_serialized == b_serialized else "!=",
                b_serialized,
            )
        )
    if a_representation != b_representation:
        message.append(
            "\n        %s\n        !=\n        %s"
            % (a_representation, b_representation)
        )
    return "\n".join(message)


def _get_object_discrepancies_error_message(
    object_a: abc.Object, object_b: abc.Object
) -> str:
    a_serialized: str = serialize(object_a)
    b_serialized: str = serialize(object_b)
    a_representation: str = "".join(
        line.strip() for line in repr(object_a).split("\n")
    )
    b_representation = "".join(
        line.strip() for line in repr(object_b).split("\n")
    )
    class_name: str = qualified_name(type(object_a))
    message = [
        "Discrepancies were found between the instance of "
        f"`{class_name}` provided and "
        "a serialized/deserialized clone:"
    ]
    if a_serialized != b_serialized or (a_representation != b_representation):
        message.append(
            "        %s\n        %s\n        %s"
            % (
                a_serialized,
                "==" if a_serialized == b_serialized else "!=",
                b_serialized,
            )
        )
    if a_representation != b_representation:
        message.append(
            "\n        %s\n        !=\n        %s"
            % (a_representation, b_representation)
        )
    property_name: str
    property_values: Tuple[
        Optional[abc.MarshallableTypes], Optional[abc.MarshallableTypes]
    ]
    for property_name, property_values in _object_discrepancies(
        object_a, object_b
    ).items():
        property_value_a: Optional[abc.MarshallableTypes]
        property_value_b: Optional[abc.MarshallableTypes]
        property_value_a, property_value_b = property_values
        message.append(
            _get_value_discrepancies_error_message(
                class_name, property_name, property_value_a, property_value_b
            )
        )
    return "\n".join(message)


def _get_object_discrepancies_error(
    object_instance: abc.Object, reloaded_object_instance: abc.Object
) -> ObjectDiscrepancyError:
    message: str = _get_object_discrepancies_error_message(
        object_instance, reloaded_object_instance
    )
    return ObjectDiscrepancyError(message)


def _remarshal_object(
    string_object: str, object_instance: abc.Object, format_: str = "json"
) -> None:
    if format_ == "yaml":
        reloaded_marshalled_data = yaml_.load(string_object)
    else:
        reloaded_marshalled_data = _json.loads(
            string_object,
            object_hook=collections.OrderedDict,
            object_pairs_hook=collections.OrderedDict,
        )
    keys: Set[str] = set()
    instance_meta: Optional[abc.Meta] = meta.read(object_instance)
    assert isinstance(instance_meta, (abc.ObjectMeta, NoneType))
    if (instance_meta is not None) and (instance_meta.properties is not None):
        for property_name, property_ in instance_meta.properties.items():
            keys.add(property_.name or property_name)
        for key in reloaded_marshalled_data.keys():
            if key not in keys:
                raise KeyError(
                    '"%s" not found in serialized/re-deserialized data: %s'
                    % (key, string_object)
                )


def _reload_object(object_instance: abc.Object, format_: str) -> None:
    object_type: type = type(object_instance)
    string_object: str = str(object_instance)
    assert string_object != ""
    reloaded_object_instance = object_type(string_object)
    meta.copy_to(object_instance, reloaded_object_instance)
    if object_instance != reloaded_object_instance:
        raise _get_object_discrepancies_error(
            object_instance, reloaded_object_instance
        )
    reloaded_string = str(reloaded_object_instance)
    if string_object != reloaded_string:
        raise ObjectDiscrepancyError(
            "\n%s\n!=\n%s" % (string_object, reloaded_string)
        )
    _remarshal_object(string_object, object_instance, format_)


def _object(
    object_instance: abc.Object,
    format_: str,
    raise_validation_errors: bool = True,
) -> None:
    errors: Sequence[str] = validate(
        object_instance, raise_errors=raise_validation_errors
    )
    if errors:
        warn("\n" + "\n".join(errors))
    _reload_object(object_instance, format_)
    instance_meta: Optional[abc.Meta] = meta.read(object_instance)
    assert isinstance(instance_meta, (abc.ObjectMeta, NoneType))
    if (instance_meta is not None) and (instance_meta.properties is not None):
        # Recursively test property values
        for property_name, property_ in instance_meta.properties.items():
            property_value = getattr(object_instance, property_name)
            if isinstance(property_value, abc.Model):
                model(
                    property_value,
                    format_=format_,
                    raise_validation_errors=raise_validation_errors,
                )


def model(
    model_instance: abc.Model,
    format_: str = "json",
    raise_validation_errors: bool = True,
) -> None:
    """
    Tests an instance of a `sob.model.Model` sub-class.

    Parameters:

        - model_instance (sob.model.Model):

            An instance of a `sob.model.Model` sub-class.

        - format_ (str):

            The serialization format being tested: 'json' or 'yaml'.

        - raise_validation_errors (bool):

            The function `sob.model.validate` verifies that all required
            attributes are present, as well as any additional validations
            implemented using the model's validation hooks `after_validate`
            and/or `before_validate`.

                - If `True`, errors resulting from `sob.model.validate` are
                  raised.

                - If `False`, errors resulting from `sob.model.validate` are
                  expressed only as warnings.
    """
    assert_is_instance("model_instance", model_instance, abc.Model)
    assert_in("format_", format_, ("json", "yaml"))
    meta.format_(model_instance, format_)
    if isinstance(model_instance, abc.Object):
        _object(model_instance, format_, raise_validation_errors)
    elif isinstance(model_instance, abc.Array):
        validate(model_instance)
        for item in model_instance:
            if isinstance(item, abc.Model):
                model(
                    item,
                    format_=format_,
                    raise_validation_errors=raise_validation_errors,
                )
    elif isinstance(model_instance, abc.Dictionary):
        validate(model_instance)
        key: str
        value: abc.MarshallableTypes
        for key, value in model_instance.items():
            if isinstance(value, abc.Model):
                model(
                    value,
                    format_=format_,
                    raise_validation_errors=raise_validation_errors,
                )


def json(
    model_instance: abc.Model,
    raise_validation_errors: bool = True,
) -> None:
    model(
        model_instance=model_instance,
        format_="json",
        raise_validation_errors=raise_validation_errors,
    )


def yaml(
    model_instance: abc.Model,
    raise_validation_errors: bool = True,
) -> None:
    model(
        model_instance=model_instance,
        format_="yaml",
        raise_validation_errors=raise_validation_errors,
    )



"""
This module defines classes for describing properties of a model.
"""

import collections
import collections.abc
from copy import deepcopy
from datetime import date, datetime
from typing import (
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
    Any,
    Callable,
)
from decimal import Decimal

from iso8601 import parse_date  # type: ignore
from itertools import chain

from . import abc
from .types import Types, MutableTypes
from .utilities import (
    indent,
    parameters_defaults,
    properties_values,
    qualified_name,
)
from .utilities.assertion import assert_is_instance
from .utilities.inspect import represent
from .utilities.types import UNDEFINED, Undefined, NoneType
from .abc import MarshallableTypes
from .version import Version

__all__: List[str] = [
    "Property",
    "Array",
    "Boolean",
    "Bytes",
    "Date",
    "Dictionary",
    "Enumerated",
    "Integer",
    "Number",
    "String",
    "TYPES_PROPERTIES",
    "has_mutable_types",
]


def _repr_keyword_argument_assignment(
    argument: str,
    value: MarshallableTypes,
    defaults: Optional[Dict[str, MarshallableTypes]] = None,
) -> Optional[str]:
    """
    Returns a string representation of an argument assignment, or `None`
    if the argument value is equal to the default value for that argument
    """
    if (defaults is not None) and (
        (argument not in defaults)
        or defaults[argument] == value
        or value is None
    ):
        return None
    return "    %s=%s," % (argument, indent(represent(value)))


def has_mutable_types(property: Union[abc.Property, type]) -> bool:
    """
    This function returns `True` if modification of the `.types` member of a
    property class or instance is permitted.

    Parameters:

    - property (sob.properties.Property|type)
    """
    property_type: type
    if isinstance(property, abc.Property):
        property_type = type(property)
    else:
        assert issubclass(property, abc.Property)
        property_type = property
    return getattr(property_type, "_types") is None


class Property(abc.Property):
    """
    This is the base class for defining a property.

    Properties

        - types ([type|Property]): One or more expected `type` or
          `Property` instances. A list of more than one types and/or properties
          results in a polymorphic interpretation wherein a value is
          un-marshalled in accordance with each type or property in the list
          (sequentially), until the value is un-marshalled without throwing a
          `TypeError` or `ValueError`. If the list of types and/or properties
          is exhausted without successfully un-marshalling the value, a
          `TypeError` or `ValueError` error is raised.

        - required (bool): If `True`—marshalling a value for this property
          will throw an error if the value is `None`. Please note that `None`
          indicates a value was *not provided*. To indicate an *explicit* null
          value, use `sob.properties.types.NULL`.

        - versions ([str]|{str:Property}):

          The parameter should be one of the following:

            - A `set`, `tuple`, or `list` of version numbers to which this
              property applies.
            - A mapping of version numbers to an instance of
              [Property](#Property) instances applicable to that version.

          Version numbers prefixed by "<" indicating any version less than the
          one specified, so "<3.0" indicates that this property is available in
          versions prior to 3.0. The inverse is true for version numbers
          prefixed by ">". ">=" and "<=" have similar meanings, but are
          inclusive.

          Versioning can be applied to a property by calling
          `sob.meta.set_version` in the `__init__` method of an
          `sob.model.Object` sub-class.

        - name (str): The name of the property when loaded from or dumped into
          a JSON object. Specifying a `name` facilitates mapping of PEP8
          compliant property names to JSON or YAML attribute names which might
          be incompatible with well-formatted python code due to various
          reasons such as being camelCased, or being python keywords. To
          infer an appropriate property name programmatically, use the utility
          function `sob.utilities.string.property_name`.
    """

    _types: Optional[abc.Types] = None

    # noinspection PyShadowingNames
    def __init__(
        self,
        types: Union[
            abc.Types,
            Sequence[Union[type, "Property"]],
            type,
            "Property",
            Undefined,
            None,
        ] = UNDEFINED,
        name: Optional[str] = None,
        required: bool = False,
        versions: Optional[
            Union[str, abc.Version, Iterable[Union[str, abc.Version]]]
        ] = None,
    ) -> None:
        self._types: Optional[abc.Types] = getattr(type(self), "_types")
        if types is not UNDEFINED:
            setattr(self, "types", types)
        self.name: Optional[str] = name
        self.required: bool = required
        self._versions: Optional[Sequence[abc.Version]] = None
        if versions is not None:
            setattr(self, "versions", versions)

    @property  # type: ignore
    def types(self) -> Optional[abc.Types]:
        return self._types

    @types.setter
    def types(
        self,
        types_or_properties: Union[
            abc.Types,
            Sequence[Union[type, abc.Property]],
            type,
            abc.Property,
            None,
        ],
    ) -> None:
        # If types are set at the class-level, don't touch them
        if type(self)._types is not None:
            raise TypeError(
                f"`{qualified_name(type(self))}.types` is immutable"
            )
        if (types_or_properties is not None) and not isinstance(
            types_or_properties, abc.Types
        ):
            types_or_properties = MutableTypes(types_or_properties)
        assert (types_or_properties is None) or isinstance(
            types_or_properties, abc.Types
        )
        self._types = types_or_properties

    @property  # type: ignore
    def versions(self) -> Optional[Sequence[abc.Version]]:
        return self._versions

    @versions.setter
    def versions(
        self,
        versions: Optional[
            Union[str, abc.Version, Iterable[Union[str, abc.Version]]]
        ] = None,
    ) -> None:
        versions_tuple: Optional[Tuple[abc.Version, ...]] = None
        if versions is not None:
            assert_is_instance(
                "versions",
                versions,
                (str, abc.Version, collections.abc.Iterable),
            )
            version: Union[str, abc.Version]
            if isinstance(versions, str):
                version = Version(versions)
                versions_tuple = (version,)
            elif isinstance(versions, abc.Version):
                versions_tuple = (versions,)
            else:
                versions_list: List[abc.Version] = []
                for version in versions:
                    if not isinstance(version, abc.Version):
                        version = Version(version)
                    versions_list.append(version)
                versions_tuple = tuple(versions_list)
        self._versions = versions_tuple

    def __repr__(self) -> str:
        lines = [qualified_name(type(self)) + "("]
        defaults: Dict[str, Any] = parameters_defaults(
            getattr(self, "__init__")
        )
        for property_name, value in properties_values(self):
            argument_representation = _repr_keyword_argument_assignment(
                property_name, value, defaults
            )
            if argument_representation is not None:
                lines.append(argument_representation)
        lines[-1] = lines[-1].rstrip(",")
        lines.append(")")
        if len(lines) > 2:
            return "\n".join(lines)
        else:
            return "".join(lines)

    def __copy__(self) -> abc.Property:
        new_instance = self.__class__()
        attribute_name: str
        for attribute_name in dir(self):
            if attribute_name[0] != "_" and attribute_name != "data":
                value = getattr(self, attribute_name)
                if not callable(value):
                    setattr(new_instance, attribute_name, value)
        return new_instance

    def __deepcopy__(self, memo: dict) -> abc.Property:
        new_instance: abc.Property = self.__class__()
        attribute: str
        value: Any
        types_is_mutable: bool = bool(type(self)._types is None)
        for attribute, value in properties_values(self):
            if attribute != "types" or types_is_mutable:
                setattr(new_instance, attribute, deepcopy(value, memo=memo))
        return new_instance


class String(Property, abc.String):
    """
    See `sob.properties.Property`
    """

    _types: abc.Types = Types((str,))  # type: ignore

    def __init__(
        self,
        name: Optional[str] = None,
        required: bool = False,
        versions: Optional[
            Union[str, abc.Version, Iterable[Union[str, abc.Version]]]
        ] = None,
    ) -> None:
        super().__init__(
            name=name,
            required=required,
            versions=versions,
        )


def _date2str(value: date) -> str:
    return value.isoformat()


class Date(Property, abc.Date):
    """
    ...See `sob.properties.Property`

    + Parameters:

        - date2str (collections.Callable): A function, taking one argument (a
          python `date` json_object), and returning a date string in the
          desired format. The default is `date.isoformat`--returning an
          iso8601 compliant date string.

        - str2date (collections.Callable): A function, taking one argument (a
          date string), and returning a python `date` object. By default,
          this is `iso8601.parse_date`.
    """

    _types: Optional[abc.Types] = Types((date,))

    def __init__(
        self,
        name: Optional[str] = None,
        required: bool = False,
        versions: Optional[
            Union[str, abc.Version, Iterable[Union[str, abc.Version]]]
        ] = None,
        date2str: Callable[[date], str] = _date2str,
        str2date: Callable[[str], date] = parse_date,
    ) -> None:
        super().__init__(
            name=name,
            required=required,
            versions=versions,
        )
        self._date2str = date2str
        self._str2date = str2date

    def date2str(self, value: date) -> str:
        return self._date2str(value)

    def str2date(self, value: str) -> date:
        return self._str2date(value)


def _datetime2str(value: datetime) -> str:
    return value.isoformat()


class DateTime(Property, abc.DateTime):
    """
    (See [`sob.properties.Property`](#Property))

    + Parameters:

    - datetime2str (collections.Callable): A function, taking one argument
      (a python `datetime` json_object), and returning a date-time string
      in the desired format. The default is `datetime.isoformat`--returning
      an iso8601 compliant date-time string.
    - str2datetime (collections.Callable): A function, taking one argument
      (a datetime string), and returning a python `datetime` json_object.
      By default, this is `iso8601.parse_date`.
    """

    _types: abc.Types = Types((datetime,))  # type: ignore

    def __init__(
        self,
        name: Optional[str] = None,
        required: bool = False,
        versions: Optional[
            Union[str, abc.Version, Iterable[Union[str, abc.Version]]]
        ] = None,
        datetime2str: Callable[[datetime], str] = _datetime2str,
        str2datetime: Callable[[str], datetime] = parse_date,
    ) -> None:
        self._datetime2str = datetime2str
        self._str2datetime = str2datetime
        super().__init__(
            name=name,
            required=required,
            versions=versions,
        )

    def datetime2str(self, value: datetime) -> str:
        return self._datetime2str(value)

    def str2datetime(self, value: str) -> datetime:
        return self._str2datetime(value)


class Bytes(Property, abc.Bytes):
    """
    (See [`sob.properties.Property`](#Property))

    This class represents a property with binary values
    """

    _types: abc.Types = Types((bytes,))  # type: ignore

    def __init__(
        self,
        name: Optional[str] = None,
        required: bool = False,
        versions: Optional[
            Union[str, abc.Version, Iterable[Union[str, abc.Version]]]
        ] = None,
    ) -> None:
        super().__init__(
            name=name,
            required=required,
            versions=versions,
        )


class Enumerated(Property, abc.Enumerated):
    """
    Parameters:

    This class accepts the following keyword parameters in *addition* to all
    parameters applicable to the base class [Property](#Property).

    - values ([typing.Any]):  A list or set of possible values.

    Properties:

    This class exposes public properties matching its keyword parameters.
    """

    # noinspection PyShadowingNames
    def __init__(
        self,
        types: Union[
            abc.Types,
            Sequence[Union[type, "Property"]],
            type,
            "Property",
            Undefined,
            None,
        ] = UNDEFINED,
        values: Optional[Iterable[MarshallableTypes]] = None,
        name: Optional[str] = None,
        required: bool = False,
        versions: Optional[
            Union[str, abc.Version, Iterable[Union[str, abc.Version]]]
        ] = None,
    ) -> None:
        self._values: Optional[Set[MarshallableTypes]] = None
        super().__init__(
            types=types, name=name, required=required, versions=versions
        )
        setattr(self, "values", values)

    @property  # type: ignore
    def values(self) -> Optional[Set[MarshallableTypes]]:
        return self._values

    @values.setter
    def values(self, values: Optional[Iterable[MarshallableTypes]]) -> None:
        if values is None:
            self._values = None
        else:
            assert_is_instance("values", values, collections.abc.Iterable)
            self._values = set(values)


class Number(Property, abc.Number):
    """
    See `sob.properties.Property`
    """

    _types: abc.Types = Types((Decimal, float, int))  # type: ignore

    def __init__(
        self,
        name: Optional[str] = None,
        required: bool = False,
        versions: Optional[
            Union[str, abc.Version, Iterable[Union[str, abc.Version]]]
        ] = None,
    ) -> None:
        super().__init__(name=name, required=required, versions=versions)


class Integer(Property, abc.Integer):
    """
    See `sob.properties.Property`
    """

    _types: abc.Types = Types((int,))  # type: ignore

    def __init__(
        self,
        name: Optional[str] = None,
        required: bool = False,
        versions: Optional[
            Union[str, abc.Version, Iterable[Union[str, abc.Version]]]
        ] = None,
    ) -> None:
        super().__init__(
            name=name,
            required=required,
            versions=versions,
        )


class Boolean(Property, abc.Boolean):
    """
    See `sob.properties.Property`
    """

    _types: abc.Types = Types((bool,))  # type: ignore

    def __init__(
        self,
        name: Optional[str] = None,
        required: bool = False,
        versions: Optional[
            Union[str, abc.Version, Iterable[Union[str, abc.Version]]]
        ] = None,
    ) -> None:
        super().__init__(
            name=name,
            required=required,
            versions=versions,
        )


class Array(Property, abc.ArrayProperty):
    """
    See `sob.properties.Property`...

    + Properties:

    - item_types (type|Property|[type|Property]): The type(s) of values/objects
      contained in the array. Similar to
      `sob.properties.Property().types`, but applied to items in the
      array, not the array itself.
    """

    _types: abc.Types = Types((abc.Array,))  # type: ignore
    _item_types: Optional[abc.Types] = None

    def __init__(
        self,
        item_types: Union[
            type, Sequence[Union[type, Property]], Undefined, abc.Types, None
        ] = UNDEFINED,
        name: Optional[str] = None,
        required: bool = False,
        versions: Optional[
            Union[str, abc.Version, Iterable[Union[str, abc.Version]]]
        ] = None,
    ) -> None:
        self._item_types: Optional[Types] = getattr(type(self), "_item_types")
        if item_types is not UNDEFINED:
            self.item_types = item_types  # type: ignore
        super().__init__(
            name=name,
            required=required,
            versions=versions,
        )

    @property  # type: ignore
    def item_types(self) -> Optional[abc.Types]:  # type: ignore
        return self._item_types

    @item_types.setter  # type: ignore
    def item_types(
        self,
        item_types: Union[
            type,
            abc.Property,
            Sequence[Union[type, abc.Property]],
            abc.Types,
            None,
        ],
    ) -> None:
        if isinstance(item_types, abc.Types):
            if not isinstance(item_types, abc.MutableTypes):
                item_types = MutableTypes(item_types)
        elif item_types is not None:
            item_types_list: List[Union[type, abc.Property]] = []
            if isinstance(item_types, (type, abc.Property)):
                item_types_list.append(item_types)
            else:
                assert isinstance(item_types, Sequence)
                for item_type in item_types:
                    assert isinstance(item_type, (type, abc.Property))
                    item_types_list.append(item_type)
            item_types = MutableTypes(item_types_list)
        assert isinstance(item_types, (abc.Types, NoneType))
        self._item_types = item_types


class Dictionary(Property, abc.DictionaryProperty):
    """
    See `sob.properties.Property`...

    + Properties:

    - value_types (type|Property|[type|Property]): The type(s) of
      values/objects comprising the mapped values. Similar to
      `sob.properties.Property.types`, but applies to *values* in the
      dictionary object, not the dictionary itself.
    """

    _types: abc.Types = Types((abc.Dictionary,))  # type: ignore
    _value_types: Optional[abc.Types] = None

    def __init__(
        self,
        value_types: Optional[
            Union[type, Sequence[Union[type, Property]], Undefined]
        ] = UNDEFINED,
        name: Optional[str] = None,
        required: bool = False,
        versions: Optional[
            Union[str, abc.Version, Iterable[Union[str, abc.Version]]]
        ] = None,
    ) -> None:
        self._value_types: Optional[abc.Types] = getattr(
            type(self), "_value_types"
        )
        if value_types is not UNDEFINED:
            self.value_types = value_types  # type: ignore
        super().__init__(
            name=name,
            required=required,
            versions=versions,
        )

    @property  # type: ignore
    def value_types(self) -> Optional[abc.Types]:
        return self._value_types

    @value_types.setter
    def value_types(
        self,
        value_types: Union[
            Sequence[Union[type, abc.Property]], abc.Types, None
        ],
    ) -> None:
        """
        A sequence of types and/or `sob.properties.Property` instances.

        If more than one type or property definition is provided,
        un-marshalling is attempted using each `value_type`, in sequential
        order. If a value could be cast into more than one of the `types`
        without throwing a `ValueError`, `TypeError`, or
        `sob.errors.ValidationError`, the value type occurring *first* in the
        sequence will be used.
        """
        if (value_types is not None) and not isinstance(
            value_types, abc.MutableTypes
        ):
            value_types = MutableTypes(value_types)
        assert isinstance(value_types, (abc.Types, NoneType))
        self._value_types = value_types


# This constant maps data types to their corresponding properties
TYPES_PROPERTIES: Dict[type, type] = {
    type_: property_class
    for type_, property_class in chain(
        *map(
            lambda property_class: (
                (type_, property_class)
                for type_ in getattr(property_class, "_types")
            ),
            (
                property_class
                for property_class in locals().values()
                if (
                    isinstance(property_class, type)
                    and issubclass(property_class, Property)
                    and getattr(property_class, "_types")
                )
            ),
        )
    )
}



from copy import deepcopy
from typing import Any, Callable, Optional, Tuple, Union

from . import abc
from .utilities.inspect import qualified_name
from .utilities.assertion import assert_is_instance
from .abc import JSONTypes, MarshallableTypes


class Hooks(abc.Hooks):
    """
    TODO
    """

    def __init__(
        self,
        before_marshal: Optional[Callable[[abc.Model], abc.Model]] = None,
        after_marshal: Optional[Callable[[JSONTypes], JSONTypes]] = None,
        before_unmarshal: Optional[
            Callable[[MarshallableTypes], MarshallableTypes]
        ] = None,
        after_unmarshal: Optional[Callable[[abc.Model], abc.Model]] = None,
        before_serialize: Optional[Callable[[JSONTypes], JSONTypes]] = None,
        after_serialize: Optional[Callable[[str], str]] = None,
        before_validate: Optional[Callable[[abc.Model], abc.Model]] = None,
        after_validate: Optional[Callable[[abc.Model], None]] = None,
    ):
        self.before_marshal = before_marshal
        self.after_marshal = after_marshal
        self.before_unmarshal = before_unmarshal
        self.after_unmarshal = after_unmarshal
        self.before_serialize = before_serialize
        self.after_serialize = after_serialize
        self.before_validate = before_validate
        self.after_validate = after_validate

    def __copy__(self) -> "Hooks":
        return self.__class__(**vars(self))

    def __deepcopy__(self, memo: dict = None) -> "Hooks":
        return self.__class__(
            **{
                key: deepcopy(value, memo=memo)
                for key, value in vars(self).items()
            }
        )

    def __bool__(self) -> bool:
        return True


class Object(Hooks, abc.ObjectHooks):
    """
    TODO
    """

    def __init__(
        self,
        before_marshal: Optional[Callable[[abc.Model], abc.Model]] = None,
        after_marshal: Optional[Callable[[JSONTypes], JSONTypes]] = None,
        before_unmarshal: Optional[
            Callable[[MarshallableTypes], MarshallableTypes]
        ] = None,
        after_unmarshal: Optional[Callable[[abc.Model], abc.Model]] = None,
        before_serialize: Optional[Callable[[JSONTypes], JSONTypes]] = None,
        after_serialize: Optional[Callable[[str], str]] = None,
        before_validate: Optional[Callable[[abc.Model], abc.Model]] = None,
        after_validate: Optional[Callable[[abc.Model], None]] = None,
        before_setattr: Optional[
            Callable[
                [abc.Object, str, MarshallableTypes],
                Tuple[str, MarshallableTypes],
            ]
        ] = None,
        after_setattr: Optional[
            Callable[[abc.Object, str, MarshallableTypes], None]
        ] = None,
        before_setitem: Optional[
            Callable[
                [abc.Object, str, MarshallableTypes],
                Tuple[str, MarshallableTypes],
            ]
        ] = None,
        after_setitem: Optional[
            Callable[[abc.Object, str, MarshallableTypes], None]
        ] = None,
    ) -> None:
        super().__init__(
            before_marshal=before_marshal,
            after_marshal=after_marshal,
            before_unmarshal=before_unmarshal,
            after_unmarshal=after_unmarshal,
            before_serialize=before_serialize,
            after_serialize=after_serialize,
            before_validate=before_validate,
            after_validate=after_validate,
        )
        self.before_setattr = before_setattr
        self.after_setattr = after_setattr
        self.before_setitem = before_setitem
        self.after_setitem = after_setitem


class Array(Hooks, abc.ArrayHooks):
    """
    TODO
    """

    def __init__(
        self,
        before_marshal: Optional[Callable[[abc.Model], abc.Model]] = None,
        after_marshal: Optional[Callable[[JSONTypes], JSONTypes]] = None,
        before_unmarshal: Optional[
            Callable[[MarshallableTypes], MarshallableTypes]
        ] = None,
        after_unmarshal: Optional[Callable[[abc.Model], abc.Model]] = None,
        before_serialize: Optional[Callable[[JSONTypes], JSONTypes]] = None,
        after_serialize: Optional[Callable[[str], str]] = None,
        before_validate: Optional[Callable[[abc.Model], abc.Model]] = None,
        after_validate: Optional[Callable[[abc.Model], None]] = None,
        before_setitem: Optional[
            Callable[
                [abc.Array, int, MarshallableTypes],
                Tuple[int, MarshallableTypes],
            ]
        ] = None,
        after_setitem: Optional[
            Callable[[abc.Array, int, MarshallableTypes], None]
        ] = None,
        before_append: Optional[
            Callable[[abc.Array, MarshallableTypes], MarshallableTypes]
        ] = None,
        after_append: Optional[
            Callable[[abc.Array, MarshallableTypes], None]
        ] = None,
    ) -> None:
        super().__init__(
            before_marshal=before_marshal,
            after_marshal=after_marshal,
            before_unmarshal=before_unmarshal,
            after_unmarshal=after_unmarshal,
            before_serialize=before_serialize,
            after_serialize=after_serialize,
            before_validate=before_validate,
            after_validate=after_validate,
        )
        self.before_setitem = before_setitem
        self.after_setitem = after_setitem
        self.before_append = before_append
        self.after_append = after_append


class Dictionary(Hooks, abc.DictionaryHooks):
    """
    TODO
    """

    def __init__(
        self,
        before_marshal: Optional[Callable[[abc.Model], abc.Model]] = None,
        after_marshal: Optional[Callable[[JSONTypes], JSONTypes]] = None,
        before_unmarshal: Optional[
            Callable[[MarshallableTypes], MarshallableTypes]
        ] = None,
        after_unmarshal: Optional[Callable[[abc.Model], abc.Model]] = None,
        before_serialize: Optional[Callable[[JSONTypes], JSONTypes]] = None,
        after_serialize: Optional[Callable[[str], str]] = None,
        before_validate: Optional[Callable[[abc.Model], abc.Model]] = None,
        after_validate: Optional[Callable[[abc.Model], None]] = None,
        before_setitem: Optional[
            Callable[
                [abc.Dictionary, str, MarshallableTypes],
                Tuple[str, MarshallableTypes],
            ]
        ] = None,
        after_setitem: Optional[
            Callable[[abc.Dictionary, str, MarshallableTypes], None]
        ] = None,
    ):
        super().__init__(
            before_marshal=before_marshal,
            after_marshal=after_marshal,
            before_unmarshal=before_unmarshal,
            after_unmarshal=after_unmarshal,
            before_serialize=before_serialize,
            after_serialize=after_serialize,
            before_validate=before_validate,
            after_validate=after_validate,
        )
        self.before_setitem = before_setitem
        self.after_setitem = after_setitem


def read(model: Union[type, abc.Model]) -> Any:
    """
    Read metadata from a model class or instance (the returned metadata may be
    inherited, and therefore should not be written to)
    """
    hooks: Optional[abc.Hooks] = getattr(model, "_hooks")
    if isinstance(model, abc.Model) and (hooks is None):
        hooks = read(type(model))
    return hooks


def object_read(model: Union[type, abc.Object]) -> Optional[abc.ObjectHooks]:
    """
    Read metadata from an `sob.model.Object` sub-class or instance (the
    returned metadata may be inherited, and therefore should not be written
    to).
    """
    return read(model)


def array_read(model: Union[type, abc.Array]) -> Optional[abc.ArrayHooks]:
    """
    Read metadata from an `sob.model.Array` sub-class or instance (the
    returned metadata may be inherited, and therefore should not be written
    to).
    """
    return read(model)


def dictionary_read(
    model: Union[type, abc.Dictionary]
) -> Optional[abc.DictionaryHooks]:
    """
    Read metadata from an `sob.model.Dictionary` sub-class or instance (the
    returned metadata may be inherited, and therefore should not be written
    to).
    """
    return read(model)


def writable(model: Union[type, abc.Model]) -> Any:
    """
    Retrieve a metadata instance. If the instance currently inherits its
    metadata from a class or superclass, this function will copy that
    metadata and assign it directly to the model instance.
    """
    hooks: Optional[abc.Hooks] = getattr(model, "_hooks")
    writable_hooks: Optional[abc.Hooks] = None
    if isinstance(model, type):
        assert issubclass(model, abc.Model)
        if hooks is None:
            new_hooks: Optional[Hooks] = (
                Object()
                if issubclass(model, abc.Object)
                else Array()
                if issubclass(model, abc.Array)
                else Dictionary()
                if issubclass(model, abc.Dictionary)
                else Hooks()
            )
            writable_hooks = new_hooks
        else:
            for base in model.__bases__:
                base_hooks: Optional[abc.Hooks]
                try:
                    base_hooks = getattr(base, "_hooks")
                except AttributeError:
                    base_hooks = None
                if hooks and (hooks is base_hooks):
                    writable_hooks = deepcopy(hooks)
                    break
    elif isinstance(model, abc.Model):
        if hooks is None:
            writable_hooks = deepcopy(writable(type(model)))
    if writable_hooks:
        setattr(model, "_hooks", writable_hooks)
    else:
        writable_hooks = hooks
    assert writable_hooks is not None
    return writable_hooks


def object_writable(model: Union[type, abc.Object]) -> abc.ObjectHooks:
    """
    Retrieve a metadata instance. If the instance currently inherits its
    metadata from a class or superclass, this function will copy that
    metadata and assign it directly to the model instance.
    """
    return writable(model)


def array_writable(model: Union[type, abc.Array]) -> abc.ArrayHooks:
    """
    Retrieve a metadata instance. If the instance currently inherits its
    metadata from a class or superclass, this function will copy that
    metadata and assign it directly to the model instance.
    """
    return writable(model)


def dictionary_writable(
    model: Union[type, abc.Dictionary]
) -> abc.DictionaryHooks:
    """
    Retrieve a metadata instance. If the instance currently inherits its
    metadata from a class or superclass, this function will copy that
    metadata and assign it directly to the model instance.
    """
    return writable(model)


def type_(model: Union[type, abc.Model]) -> type:
    """
    Get the type of metadata required for an object
    """
    hooks_type: type
    assert_is_instance(
        "model",
        model,
        (type, abc.Object, abc.Dictionary, abc.Array),
    )
    if isinstance(model, type):
        hooks_type = (
            Object
            if issubclass(model, abc.Object)
            else Array
            if issubclass(model, abc.Array)
            else Dictionary
        )
    else:
        hooks_type = (
            Object
            if isinstance(model, abc.Object)
            else Array
            if isinstance(model, abc.Array)
            else Dictionary
        )
    return hooks_type


def write(model: Union[type, abc.Model], hooks: Optional[abc.Hooks]) -> None:
    """
    Write metadata to a class or instance
    """
    if hooks is not None:
        # Verify that the metadata is of the correct type
        hooks_type: type = type_(model)
        if not isinstance(hooks, hooks_type):
            raise ValueError(
                f"Hooks assigned to `{qualified_name(type(model))}` "
                f"must be of type `{qualified_name(hooks_type)}`"
            )
    setattr(model, "_hooks", hooks)



import os
from io import UnsupportedOperation
from typing import Callable, Optional, Union
from urllib.parse import urljoin

from .inspect import get_method
from ..abc import Readable


def read(file: Readable) -> Union[str, bytes]:
    """
    Read a file-like object and return the text or binary data it contains.

    Parameters:

    - file (abc.io.Readable): A readable, file-like object.

    This function returns an instance of `str` or `bytes`.
    """
    read_method_name: str
    seek_method: Optional[Callable] = get_method(file, "seek", None)
    if seek_method:
        try:
            seek_method(0)
        except UnsupportedOperation:
            pass
    for read_method_name in ("readall", "read"):
        read_method: Optional[Callable] = get_method(
            file, read_method_name, None
        )
        if read_method:
            try:
                return read_method()
            except UnsupportedOperation:
                pass
    raise TypeError(f"{repr(file)} is not a file-like object")


def get_url(file: Readable) -> Optional[str]:
    """
    Get the URL from which an input-output (file-like) object was sourced.

    Parameters:

    - file (abc.io.Readable)

    This function returns an instance of `str` if the originating URL or
    file-path can be inferred, otherwise it returns `None`.
    """
    url: Optional[str] = getattr(file, "url", None)
    if url is None:
        url = getattr(file, "name", None)
        if url is not None:
            url = urljoin("file:", os.path.abspath(url))
    return url



import enum
import re
import sys

import builtins
from keyword import iskeyword
from typing import List, Match, Optional, Pattern, Tuple
from unicodedata import normalize


_DIGITS: str = "0123456789"
# noinspection SpellCheckingInspection
_LOWERCASE_ALPHABET: str = "abcdefghijklmnopqrstuvwxyz"
# noinspection SpellCheckingInspection
_UPPERCASE_ALPHABET: str = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
_ALPHANUMERIC_CHARACTERS = (
    f"{_DIGITS}" f"{_UPPERCASE_ALPHABET}" f"{_LOWERCASE_ALPHABET}"
)
_URL_DIRECTORY_AND_FILE_NAME_RE: Pattern = re.compile(r"^(.*/)([^/]*)")
MAX_LINE_LENGTH: int = 79


def property_name(string: str) -> str:
    """
    Converts a "camelCased" attribute/property name, a name which conflicts
    with a python keyword, or an otherwise non-compatible string to a PEP-8
    compliant property name.

    >>> print(property_name('theBirdsAndTheBees'))
    the_birds_and_the_bees

    >>> print(property_name('FYIThisIsAnAcronym'))
    fyi_this_is_an_acronym

    >>> print(property_name('in'))
    in_

    >>> print(property_name('id'))
    id_

    >>> print(property_name('one2one'))  # No change needed
    one2one

    >>> print(property_name('One2One'))
    one_2_one

    >>> print(property_name('@One2One'))
    one_2_one
    """
    name: str = string
    # Replace accented and otherwise modified latin characters with their
    # basic latin equivalent
    name = normalize("NFKD", name)
    # Replace any remaining non-latin characters with underscores
    name = re.sub(r"([^\x20-\x7F]|\s)+", "_", name)
    # Insert underscores between lowercase and uppercase characters
    name = re.sub(r"([a-z])([A-Z])", r"\1_\2", name)
    # Insert underscores between uppercase characters and following uppercase
    # characters which are followed by lowercase characters (indicating the
    # latter uppercase character was intended as part of a capitalized word
    name = re.sub(r"([A-Z])([A-Z])([a-z])", r"\1_\2\3", name)
    # Replace any series of one or more non-alphanumeric characters remaining
    # with a single underscore
    name = re.sub(r"[^\w_]+", "_", name).lower()
    # Only insert underscores between letters and numbers if camelCasing is
    # found in the original string
    if string != string.lower() and string != string.upper():
        name = re.sub(r"([0-9])([a-zA-Z])", r"\1_\2", name)
        name = re.sub(r"([a-zA-Z])([0-9])", r"\1_\2", name)
    # Replace any two or more adjacent underscores with a single underscore
    name = re.sub(r"__+", "_", name)
    # Append an underscore to the keyword until it does not conflict with any
    # python keywords or built-ins
    while iskeyword(name) or (name in builtins.__dict__) or name == "self":
        name += "_"
    return name.lstrip("_")


def class_name(string: str) -> str:
    """
    This function accepts a string and returns a variation of that string
    which is a PEP-8 compatible python class name.

    >>> print(class_name('the birds and the bees'))
    TheBirdsAndTheBees

    >>> print(class_name('the-birds-and-the-bees'))
    TheBirdsAndTheBees

    >>> print(class_name('**the - birds - and - the - bees**'))
    TheBirdsAndTheBees

    >>> print(class_name('FYI is an acronym'))
    FYIIsAnAcronym

    >>> print(class_name('in-you-go'))
    InYouGo

    >>> print(class_name('False'))
    False_

    >>> print(class_name('True'))
    True_

    >>> print(class_name('ABC Acronym'))
    ABCAcronym

    >>> print(class_name('AB CD Efg'))
    ABCdEfg
    """
    name = camel(string, capitalize=True)
    while iskeyword(name) or (name in builtins.__dict__):
        name += "_"
    return name


def camel(string: str, capitalize: bool = False) -> str:
    """
    This function returns a camelCased representation of the input string.
    When/if an input string corresponds to a python keyword,

    Parameters:

    - string (str): The string to be camelCased.

    - capitalize (bool):

      If this is `true`, the first letter will be capitalized.

    >>> print(camel('the birds and the bees'))
    theBirdsAndTheBees

    >>> print(camel('the birds and the bees', capitalize=True))
    TheBirdsAndTheBees

    >>> print(camel('the-birds-and-the-bees'))
    theBirdsAndTheBees

    >>> print(camel('**the - birds - and - the - bees**'))
    theBirdsAndTheBees

    >>> print(camel('FYI is an acronym'))
    FYIIsAnAcronym

    >>> print(camel('in-you-go'))
    inYouGo

    >>> print(camel('False'))
    false

    >>> print(camel('True'))
    true

    >>> print(camel('in'))
    in

    >>> print(camel('AB CD Efg', capitalize=True))
    ABCdEfg

    >>> print(camel('ABC DEF GHI', capitalize=True))
    AbcDefGhi

    >>> print(camel('ABC_DEF_GHI', capitalize=True))
    AbcDefGhi

    >>> print(camel('ABC DEF GHI'))
    abcDefGhi

    >>> print(camel('ABC_DEF_GHI'))
    abcDefGhi
    """
    index: int
    character: str
    string = normalize("NFKD", string)
    characters: List[str] = []
    all_uppercase: bool = string.upper() == string
    capitalize_next: bool = capitalize
    uncapitalize_next: bool = (not capitalize) and (
        len(string) < 2
        or all_uppercase
        or not (
            string[0] in _UPPERCASE_ALPHABET
            and string[1] in _UPPERCASE_ALPHABET
        )
    )
    for index, character in enumerate(string):
        if character in _ALPHANUMERIC_CHARACTERS:
            if capitalize_next:
                if all_uppercase:
                    uncapitalize_next = True
                elif capitalize or characters:
                    character = character.upper()
                    # This prevents two acronyms which are adjacent from
                    # retaining capitalization (since word separations would
                    # not be possible to identify if caps were kept for both)
                    if characters and (characters[-1] in _UPPERCASE_ALPHABET):
                        uncapitalize_next = True
            elif uncapitalize_next:
                if character in _LOWERCASE_ALPHABET:
                    uncapitalize_next = False
                else:
                    character = character.lower()
            characters.append(character)
            capitalize_next = False
        else:
            capitalize_next = True
            uncapitalize_next = False
    character_string = "".join(characters)
    return character_string


class _CharacterType(enum.Enum):

    DIGIT = enum.auto()
    LOWERCASE = enum.auto()
    UPPERCASE = enum.auto()
    OTHER = enum.auto()


def camel_split(string: str) -> Tuple[str, ...]:
    """
    Split a string of camelCased words into a tuple.

    Examples:

    >>> print(
    ...     '(%s)' % ', '.join(
    ...         "'%s'" % s for s in camel_split('theBirdsAndTheBees')
    ...     )
    ... )
    ('the', 'Birds', 'And', 'The', 'Bees')
    >>> print(
    ...     '(%s)' % ', '.join(
    ...         "'%s'" % s for s in camel_split('theBirdsAndTheBees123')
    ...     )
    ... )
    ('the', 'Birds', 'And', 'The', 'Bees', '123')
    >>> print(
    ...     '(%s)' % ', '.join(
    ...         "'%s'" % s for s in camel_split('theBirdsAndTheBeesABC123')
    ...     )
    ... )
    ('the', 'Birds', 'And', 'The', 'Bees', 'ABC', '123')
    >>> print(
    ...     '(%s)' % ', '.join(
    ...         "'%s'" % s for s in camel_split(
    ...             'the-Birds-&-The-Bs-ABC--123'
    ...         )
    ...     )
    ... )
    ('the', '-', 'Birds', '-&-', 'The', '-', 'Bs', '-', 'ABC', '--', '123')
    >>> print(
    ...     '(%s)' % ', '.join(
    ...         "'%s'" % s for s in camel_split('THEBirdsAndTheBees')
    ...     )
    ... )
    ('THE', 'Birds', 'And', 'The', 'Bees')
    """
    words: List[List[str]] = []
    preceding_character_type: Optional[_CharacterType] = None
    for character in string:
        character_type: _CharacterType = (
            _CharacterType.LOWERCASE
            if character in _LOWERCASE_ALPHABET
            else _CharacterType.DIGIT
            if character in _DIGITS
            else _CharacterType.UPPERCASE
            if character in _UPPERCASE_ALPHABET
            else _CharacterType.OTHER
        )
        if character_type == _CharacterType.LOWERCASE:
            if preceding_character_type == _CharacterType.LOWERCASE:
                # If following another lowercase character, a lowercase
                # character always continues that word
                words[-1].append(character)
            elif preceding_character_type == _CharacterType.UPPERCASE:
                if len(words[-1]) > 1:
                    # When following a multi-character uppercase word,
                    # the preceding word's last character should be removed
                    # and a new word created from that preceding character
                    # as well as the current lowercase character (until
                    # followed by a lowercase character, the preceding
                    # uppercase character was inferred to be part of an,
                    # however now we know it was either following an acronym,
                    # or following a single-character word)
                    words.append([words[-1].pop()] + [character])
                else:
                    # When following an uppercase character, a lowercase
                    # character should be added to the preceding word if that
                    # word has only one character thus far
                    words[-1].append(character)
            else:
                words.append([character])
            preceding_character_type = _CharacterType.LOWERCASE
        else:
            # Any type of character besides one from the *lowercase alphabet*
            # should start a new word if it follows a character of a
            # different type
            if preceding_character_type == character_type:
                words[-1].append(character)
            else:
                words.append([character])
            preceding_character_type = character_type
    return tuple("".join(word) for word in words)


def indent(
    string: str,
    number_of_spaces: int = 4,
    start: int = 1,
    stop: Optional[int] = None,
) -> str:
    """
    Indent text by `number_of_spaces` starting at line index `start` and
    stopping at line index `stop`.
    """
    indented_text = string
    if ("\n" in string) or start == 0:
        lines: List[str] = string.split("\n")
        if stop:
            if stop < 0:
                stop = len(lines) - stop
        else:
            stop = len(lines)
        index: int
        for index in range(start, stop):
            line: str = lines[index]
            line_indent: str = " " * number_of_spaces
            lines[index] = f"{line_indent}{line}".rstrip()
        indented_text = "\n".join(lines)
    return indented_text


def url_directory_and_file_name(url: str) -> Tuple[str, str]:
    """
    Split a URL into a directory path and file name
    """
    directory: str
    file_name: str
    matched: Optional[Match] = _URL_DIRECTORY_AND_FILE_NAME_RE.match(url)
    assert matched is not None
    directory, file_name = matched.groups()
    return directory, file_name


def url_relative_to(absolute_url: str, base_url: str) -> str:
    """
    Returns a relative URL given an absolute URL and a base URL
    """
    # If no portion of the absolute URL is shared with the base URL--the
    # absolute URL will be returned
    relative_url: str = absolute_url
    base_url = url_directory_and_file_name(base_url)[0]
    if base_url:
        relative_url = ""
        # URLs are not case-sensitive
        base_url = base_url.lower()
        lowercase_absolute_url = absolute_url.lower()
        while base_url and (
            base_url.lower() != lowercase_absolute_url[: len(base_url)]
        ):
            relative_url = "../" + relative_url
            base_url = url_directory_and_file_name(base_url[:-1])[0]
        base_url_length: int = len(base_url)
        relative_url += absolute_url[base_url_length:]
    return relative_url


def split_long_comment_line(
    line: str, max_line_length: int = MAX_LINE_LENGTH, prefix: str = "#"
) -> str:
    """
    Split a comment (or docstring) line

    >>> print(split_long_comment_line(
    ...     '    Lorem ipsum dolor sit amet, consectetur adipiscing elit. '
    ...     'Nullam faucibu odio a urna elementum, eu tempor nisl efficitur.'
    ... ))
        Lorem ipsum dolor sit amet, consectetur adipiscing elit. Nullam faucibu
        odio a urna elementum, eu tempor nisl efficitur.
    """
    if len(line) > max_line_length:
        matched: Optional[Match] = re.match(
            (r"^[ ]*(?:%s[ ]*)?" % prefix if prefix else r"^[ ]*"), line
        )
        assert matched is not None
        indent_: str = matched.group()
        indent_length = len(indent_)
        words = re.split(r'([\w]*[\w,/"\'.;\-?`])', line[indent_length:])
        lines: List[str] = []
        wrapped_line: str = ""
        for word in words:
            if (
                len(wrapped_line) + len(word) + indent_length
            ) <= max_line_length:
                wrapped_line += word
            else:
                lines.append(indent_ + wrapped_line.rstrip())
                wrapped_line = "" if not word.strip() else word
        if wrapped_line:
            lines.append(f"{indent_}{wrapped_line}".rstrip())
        wrapped_line = "\n".join(lines)
    else:
        wrapped_line = line
    return wrapped_line


def split_long_docstring_lines(
    docstring: str, max_line_length: int = MAX_LINE_LENGTH
) -> str:
    """
    Split long docstring lines

    >>> print(split_long_docstring_lines(
    ...     '    Lorem ipsum dolor sit amet, consectetur adipiscing elit. '
    ...     'Nullam faucibu odio a urna elementum, eu tempor nisl efficitu'
    ... ))
        Lorem ipsum dolor sit amet, consectetur adipiscing elit. Nullam faucibu
        odio a urna elementum, eu tempor nisl efficitu
    """
    line: str
    indent_: str = "    "
    if "\t" in docstring:
        docstring = docstring.replace("\t", indent_)
    lines: List[str] = docstring.split("\n")
    indentation_length: int = sys.maxsize
    for line in filter(lambda line: line, lines):
        matched = re.match(r"^[ ]+", line)
        if matched:
            indentation_length = min(indentation_length, len(matched.group()))
        else:
            indentation_length = 0
            break
    if indentation_length < sys.maxsize:
        wrapped_lines: List[str] = []
        for line in lines:
            wrapped_lines.append(
                split_long_comment_line(
                    indent_ + line[indentation_length:],
                    max_line_length,
                    prefix="",
                )
            )
        docstring = "\n".join(wrapped_lines)
    # Strip trailing whitespace and empty lines
    return re.sub(r"[ ]+(\n|$)", r"\1", docstring)


def suffix_long_lines(
    text: str, max_line_length: int = MAX_LINE_LENGTH, suffix: str = "  # noqa"
) -> str:
    """
    This function adds a suffix to the end of any line of code longer than
    `max_line_length`.

    Parameters:

    - text (str): Text representing python code
    - max_line_length (int) = 79:
      The length at which a line should have the `suffix` appended. If
      this is a *negative* integer (or zero), the sum of this integer +
      `MAX_LINE_LENGTH` is used
    - suffix (str) = "  # noqa": The default suffix indicates to linters that
      a long line should be permitted
    """
    if max_line_length <= 0:
        max_line_length += MAX_LINE_LENGTH

    def suffix_line_if_long(line: str) -> str:
        if len(line) > max_line_length and not line.endswith(suffix):
            line = f"{line}{suffix}"
        return line

    return "\n".join(map(suffix_line_if_long, text.split("\n")))



"""
This module extends the functionality of `urllib.request.Request` to support
multipart requests, to support passing instances of sob models to the `data`
parameter/property for `urllib.request.Request`, and to support casting
requests as `str` or `bytes` (typically for debugging purposes and/or to aid in
producing non-language-specific API documentation).
"""
import collections
import collections.abc
import random
import re
import string
from typing import (
    Dict,
    ItemsView,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
)
from urllib.request import Request as _Request  # type: ignore

from . import abc
from .model import Dictionary, serialize


class Headers(collections.OrderedDict):
    """
    A dictionary of headers for a `Request`, `Part`, or `MultipartRequest`
    instance.
    """

    def __init__(
        self,
        items: Union[Dict[str, str], Iterable[Tuple[str, str]]],
        request: "Data",
    ) -> None:
        assert isinstance(request, Data)
        self.request: Data = request
        super().__init__(items)

    def _reset_part(self) -> None:
        if isinstance(self.request, Part):
            del self.request.boundary
            self.request.clear_bytes()

    def pop(  # type: ignore
        self, key: str, default: Optional[str] = None
    ) -> Optional[str]:
        key = key.capitalize()
        self._reset_part()
        return super().pop(key, default)

    def popitem(self, last: bool = True) -> Tuple[str, str]:
        self._reset_part()
        return super().popitem(last=last)

    def setdefault(self, key: str, default: Optional[str] = None) -> str:
        key = key.capitalize()
        self._reset_part()
        return super().setdefault(key, default)

    def update(
        self,
        *args: Union[
            Mapping[str, str], abc.Dictionary, Iterable[Tuple[str, str]]
        ],
        **kwargs: str,
    ) -> None:
        capitalized_dict: abc.OrderedDict[str, str] = collections.OrderedDict()
        key: str
        value: str
        for key, value in collections.OrderedDict(*args, **kwargs).items():
            capitalized_dict[key.capitalize()] = value
        self._reset_part()
        super().update(capitalized_dict)

    def __getitem__(self, key: str) -> str:
        key = key.capitalize()
        value: Optional[str]
        if key == "Content-length":
            value = str(self._get_content_length())
        elif key == "Content-type":
            value = self._get_content_type()
        else:
            value = super().__getitem__(key)
        assert isinstance(value, str)
        return value

    def __delitem__(self, key: str) -> None:
        self._reset_part()
        super().__delitem__(key.capitalize())

    def __setitem__(self, key: str, value: str) -> None:
        key = key.capitalize()
        if key != "Content-length":
            self._reset_part()
            super().__setitem__(key, value)

    def _get_content_length(self) -> int:
        return len(self.request.data or b"")

    def _get_boundary(self) -> str:
        boundary: str = ""
        if isinstance(self.request, Part):
            boundary = str(self.request.boundary, encoding="utf-8")
        return boundary

    def _get_content_type(self) -> str:
        if isinstance(self.request, Part) and self.request.parts:
            return f"multipart/form-data; boundary={self._get_boundary()}"
        else:
            return super().__getitem__("Content-type")

    def __len__(self) -> int:
        return super().__len__()

    def __iter__(self) -> Iterator[str]:
        keys: Set[str] = set()
        key: str
        for key in super().__iter__():
            keys.add(key)
            yield key
        if type(self.request) is not Part:
            # *Always* include "Content-length"
            if "Content-length" not in keys:
                yield "Content-length"
        if isinstance(self.request, Part) and self.request.parts:
            # Always include "Content-type" for multi-part requests
            if "Content-type" not in keys:
                yield "Content-type"

    def __contains__(self, key: str) -> bool:  # type: ignore
        return super().__contains__(key.capitalize())

    def items(self) -> ItemsView[str, str]:  # type: ignore
        for key in self.__iter__():
            yield key, self[key]

    def keys(self) -> Iterator[str]:  # type: ignore
        key: str
        for key in self.__iter__():
            yield key

    def values(self) -> Iterator[str]:  # type: ignore
        key: str
        for key in self.__iter__():
            yield self[key]

    def copy(self) -> "Headers":
        return self.__class__(super().items(), request=self.request)

    def __copy__(self) -> "Headers":
        return self.copy()


class Data:
    """
    One of a multipart request's parts.
    """

    def __init__(
        self,
        data: Optional[
            Union[bytes, str, Sequence, Set, dict, abc.Model]
        ] = None,
        headers: Optional[Dict[str, str]] = None,
    ) -> None:
        """
        Parameters:

            - data (
                  bytes|str|collections.Sequence|collections.Set|dict|
                  abc.model.Model
              ): The payload.
            - headers ({str: str}): A dictionary of headers (for this part of
              the request body, not the main request). This should (almost)
              always include values for "Content-Disposition" and
              "Content-Type".
        """
        self._bytes: Optional[bytes] = None
        self._headers: Optional[Headers] = None
        self._data: Optional[bytes] = None
        setattr(self, "headers", headers)
        setattr(self, "data", data)

    @property  # type: ignore
    def headers(self) -> Optional[Headers]:
        return self._headers

    @headers.setter
    def headers(self, headers: Union[Dict[str, str], Headers, None]) -> None:
        self._bytes = None
        if headers is None:
            headers = Headers({}, self)
        elif isinstance(headers, Headers):
            headers.request = self
        else:
            headers = Headers(headers, self)
        self._headers = headers

    @property  # type: ignore
    def data(self) -> Optional[bytes]:  # type: ignore
        return self._data

    @data.setter
    def data(
        self,
        data: Optional[Union[bytes, str, Sequence, Set, dict, abc.Model]],
    ) -> None:
        self._bytes = None
        if data is not None:
            serialize_type = None
            if (self.headers is not None) and ("Content-type" in self.headers):
                content_type: str = self.headers["Content-type"]
                if re.search(r"/json\b", content_type) is not None:
                    serialize_type = "json"
                if re.search(r"/yaml\b", content_type) is not None:
                    serialize_type = "yaml"
            if isinstance(data, (abc.Model, dict)) or (
                isinstance(
                    data, (collections.abc.Sequence, collections.abc.Set)
                )
                and not isinstance(data, (str, bytes))
            ):
                data = serialize(data, serialize_type or "json")
            elif isinstance(data, dict):
                data = serialize(Dictionary(data), serialize_type or "json")
            if isinstance(data, str):
                data = bytes(data, encoding="utf-8")
        self._data = data

    @data.deleter
    def data(self) -> None:  # type: ignore
        self.data = None

    def clear_bytes(self) -> None:
        self._bytes = None

    def __bytes__(self) -> bytes:
        if self._bytes is None:
            lines: List[bytes] = []
            if self.headers is not None:
                key: str
                value: str
                for key, value in self.headers.items():
                    lines.append(
                        bytes("%s: %s" % (key, value), encoding="utf-8")
                    )
            lines.append(b"")
            if self.data is not None:
                lines.append(self.data)
            self._bytes = b"\r\n".join(lines) + b"\r\n"
        return self._bytes

    def __str__(self) -> str:
        return (
            repr(bytes(self))[2:-1]
            .replace("\\r\\n", "\r\n")
            .replace("\\n", "\n")
        )


class Part(Data):
    """
    TODO
    """

    def __init__(
        self,
        data: Optional[
            Union[bytes, str, Sequence, Set, dict, abc.Model]
        ] = None,
        headers: Optional[Dict[str, str]] = None,
        parts: Optional[Sequence["Part"]] = None,
    ):
        """
        Parameters:

        - data (
              bytes|str|collections.Sequence|collections.Set|dict|
              abc.model.Model
          ): The payload.
        - headers ({str: str}): A dictionary of headers (for this part of
          the request body, not the main request). This should (almost)
          always include values for "Content-Disposition" and
          "Content-Type".
        """
        self._boundary: Optional[bytes] = None
        self._parts: Optional[Parts] = None
        self.parts = parts  # type: ignore
        Data.__init__(self, data=data, headers=headers)

    @property  # type: ignore
    def boundary(self) -> bytes:
        """
        Calculates a boundary which is not contained in any of the request
        parts.
        """
        if self._boundary is None:
            part: Part
            data: bytes = b"\r\n".join(
                [self._data or b""]
                + [bytes(part) for part in (self.parts or ())]
            )
            boundary: bytes = b"".join(
                bytes(
                    random.choice(string.digits + string.ascii_letters),
                    encoding="utf-8",
                )
                for _index in range(16)
            )
            while boundary in data:
                boundary += bytes(
                    random.choice(string.digits + string.ascii_letters),
                    encoding="utf-8",
                )
            self._boundary = boundary
        return self._boundary

    @boundary.deleter
    def boundary(self) -> None:
        self._boundary = None

    @property  # type: ignore
    def data(self) -> Optional[bytes]:
        data: Optional[bytes]
        if self.parts:
            part: Part
            data = b"%s\r\n--%s--" % (
                (b"\r\n--%s\r\n" % self.boundary).join(
                    [self._data or b""]
                    + [bytes(part).rstrip() for part in self.parts]
                ),
                self.boundary,
            )
        else:
            data = self._data
        return data

    @data.setter
    def data(
        self,
        data: Optional[Union[bytes, str, Sequence, Set, dict, abc.Model]],
    ) -> None:
        getattr(Data.data, "__set__")(self, data)

    @property  # type: ignore
    def parts(self) -> Optional["Parts"]:
        return self._parts

    @parts.setter
    def parts(self, parts: Optional[Sequence["Part"]]) -> None:
        if parts is None:
            parts = Parts([], request=self)
        elif isinstance(parts, Parts):
            parts.request = self
        else:
            parts = Parts(parts, request=self)
        self._boundary = None
        self._parts = parts


class Parts(list):
    def __init__(self, items: Sequence[Part], request: "Part") -> None:
        self.request = request
        super().__init__(items)

    def append(self, item: Part) -> None:
        self.request._boundary = None
        self.request._bytes = None
        super().append(item)

    def clear(self) -> None:
        self.request._boundary = None
        self.request._bytes = None
        super().clear()

    def extend(self, items: Iterable[Part]) -> None:
        self.request._boundary = None
        self.request._bytes = None
        super().extend(items)

    def reverse(self) -> None:
        self.request._boundary = None
        self.request._bytes = None
        super().reverse()

    def __delitem__(self, key: int) -> None:  # type: ignore
        self.request._boundary = None
        self.request._bytes = None
        super().__delitem__(key)

    def __setitem__(self, key: int, value: Part) -> None:  # type: ignore
        self.request._boundary = None
        self.request._bytes = None
        super().__setitem__(key, value)  # type: ignore


class Request(Data, _Request):  # type: ignore
    """
    A sub-class of `urllib.request.Request` which accommodates additional data
    types, and serializes `data` in accordance with what is indicated by the
    request's "Content-Type" header.
    """

    def __init__(
        self,
        url: str,
        data: Optional[
            Union[bytes, str, Sequence, Set, dict, abc.Model]
        ] = None,
        headers: Optional[Dict[str, str]] = None,
        origin_req_host: Optional[str] = None,
        unverifiable: bool = False,
        method: Optional[str] = None,
    ) -> None:
        self._bytes: Optional[bytes] = None
        self._headers = None
        self._data = None
        self.headers = headers  # type: ignore
        _Request.__init__(
            self,
            url,
            data=data,  # type: ignore
            headers=self.headers,  # type: ignore
            origin_req_host=origin_req_host,
            unverifiable=unverifiable,
            method=method,
        )


class MultipartRequest(Part, Request):  # type: ignore
    """
    A sub-class of `Request` which adds a property (and initialization
    parameter) to hold the `parts` of a multipart request.

    https://www.w3.org/Protocols/rfc1341/7_2_Multipart.html
    """

    def __init__(
        self,
        url: str,
        data: Optional[
            Union[bytes, str, Sequence, Set, dict, abc.Model]
        ] = None,
        headers: Optional[Dict[str, str]] = None,
        origin_req_host: Optional[str] = None,
        unverifiable: bool = False,
        method: Optional[str] = None,
        parts: Optional[Sequence[Part]] = None,
    ) -> None:
        Part.__init__(self, data=data, headers=headers, parts=parts)
        Request.__init__(
            self,
            url,
            data=data,
            headers=headers,
            origin_req_host=origin_req_host,
            unverifiable=unverifiable,
            method=method,
        )



"""
This module provides functionality for creating a data model from a
set of example structures.
"""
import binascii
import collections
import collections.abc
import functools
import os
from base64 import b64decode
from copy import copy, deepcopy
from datetime import date, datetime
from types import ModuleType
from typing import (
    Any,
    Callable,
    Collection,
    Dict,
    ItemsView,
    Iterable,
    Iterator,
    KeysView,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
    ValuesView,
)
from urllib.parse import quote_plus

from iso8601 import ParseError, parse_date  # type: ignore

from . import __name__ as _parent_module_name, abc, meta
from .abc import MARSHALLABLE_TYPES
from .errors import IsInstanceAssertionError
from .meta import escape_reference_token
from .model import detect_format, from_meta, unmarshal
from .properties import Property, TYPES_PROPERTIES, has_mutable_types
from .types import MutableTypes
from .utilities import get_source, qualified_name
from .utilities.assertion import (
    assert_is_instance,
    assert_is_subclass,
    assert_not_is_instance,
)
from .utilities.inspect import calling_module_name
from .utilities.io import read
from .utilities.string import (
    class_name,
    suffix_long_lines,
    property_name,
)
from .utilities.types import NULL, NoneType, Null, UNDEFINED, Undefined

__all__: List[str] = ["Synonyms", "Thesaurus"]

lru_cache: Callable[..., Any] = functools.lru_cache


def _read(data: abc.Readable) -> str:
    string_data: str
    read_data: Union[str, bytes] = read(data)
    if isinstance(read_data, (bytes, bytearray)):
        string_data = str(read_data, encoding="utf-8")
    else:
        string_data = read_data
    return string_data


def _update_types(
    types: abc.MutableTypes, new_types: abc.Types, memo: Dict[str, type]
) -> None:
    """
    This function updates `types` to incorporate any additional types from
    `new_types`, as well as to merge definitions for shared types.

    Parameters:

    - types (sob.properties.types.Types)
    - new_types (sob.properties.types.Types)
    - memo (dict)
    """
    new_type: Union[type, abc.Property]
    for new_type in new_types:
        if isinstance(new_type, type) and issubclass(new_type, abc.Model):
            if type.__name__ in memo:
                existing_type: type = memo[type.__name__]
                new_type_meta: Optional[abc.Meta] = meta.read(new_type)
                assert isinstance(
                    new_type_meta,
                    (abc.ObjectMeta, abc.ArrayMeta, abc.DictionaryMeta),
                )
                _update_model_class_from_meta(
                    existing_type, new_type_meta, memo=memo
                )
                new_type = existing_type
        if new_type not in types:
            types.append(new_type)


def _update_array_meta(
    metadata: abc.ArrayMeta,
    new_metadata: abc.ArrayMeta,
    memo: Optional[Dict[str, type]] = None,
) -> None:
    """
    This function updates `metadata` by adding/updating `metadata.item_types`
    to include types from `new_metadata.item_types`.

    Parameters:

    - metadata (sob.meta.Array)
    - new_metadata (sob.meta.Array)
    - memo (dict)
    """
    assert memo is not None
    new_item_types: Optional[abc.Types] = new_metadata.item_types
    if new_item_types is not None:
        item_types: abc.MutableTypes = (
            MutableTypes()
            if metadata.item_types is None
            else metadata.item_types
            if isinstance(metadata.item_types, abc.MutableTypes)
            else MutableTypes(metadata.item_types)
        )
        _update_types(item_types, new_item_types, memo)
        metadata.item_types = item_types  # type: ignore


def _update_dictionary_meta(
    metadata: abc.DictionaryMeta,
    new_metadata: abc.DictionaryMeta,
    memo: Optional[Dict[str, type]] = None,
) -> None:
    """
    This function updates `metadata` by adding/updating `metadata.value_types`
    to include types from `new_metadata.value_types`.

    Parameters:

    - metadata (sob.meta.Dictionary)
    - new_metadata (sob.meta.Dictionary)
    - memo (dict)
    """
    assert memo is not None
    new_value_types: Optional[abc.Types] = new_metadata.value_types
    if new_value_types is not None:
        value_types: abc.MutableTypes = (
            MutableTypes()
            if metadata.value_types is None
            else metadata.value_types
            if isinstance(metadata.value_types, abc.MutableTypes)
            else MutableTypes(metadata.value_types)
        )
        _update_types(value_types, new_value_types, memo)
        metadata.value_types = value_types  # type: ignore


def _update_object_meta(
    metadata: abc.ObjectMeta,
    new_metadata: abc.ObjectMeta,
    memo: Optional[Dict[str, type]] = None,
) -> None:
    """
    This function updates `metadata` by adding any properties from
    `new_metadata` not yet part of `metadata`, and updating the property
    types for all shared properties.

    Parameters:

    - metadata (sob.meta.Array)
    - new_metadata (sob.meta.Array)
    - memo (dict)
    """
    assert memo is not None
    assert metadata.properties is not None
    assert new_metadata.properties is not None
    metadata_keys: Set[str] = set(metadata.properties.keys())
    new_metadata_keys: Set[str] = set(new_metadata.properties.keys())
    # Add properties that don't exist
    key: str
    for key in metadata_keys - new_metadata_keys:
        metadata.properties[key] = new_metadata.properties[key]
    # Update shared properties
    for key in metadata_keys & new_metadata_keys:
        types_: Optional[abc.Types] = metadata.properties[key].types
        new_types: Optional[abc.Types] = new_metadata.properties[key].types
        if new_types is not None:
            mutable_types: abc.MutableTypes = (
                MutableTypes()
                if types_ is None
                else types_
                if isinstance(types_, abc.MutableTypes)
                else MutableTypes(types_)
            )
            _update_types(
                mutable_types,
                new_types,
                memo=memo,
            )
            types_ = mutable_types
        property: abc.Property = metadata.properties[key]
        if has_mutable_types(property):
            metadata.properties[key].types = types_  # type: ignore
        else:
            # If a property's types are immutable, we need to replace
            # it with a generic property
            metadata.properties[key] = Property(
                types=types_,
                name=property.name,
                required=property.required,
                versions=property.versions,
            )


def _update_object_class_from_meta(
    object_class: type,
    new_object_metadata: abc.ObjectMeta,
    memo: Dict[str, type],
) -> None:
    """
    This function merges new metadata into an existing object model's metadata.

    Parameters:

    - object_class (type)
    - new_object_metadata (sob.meta.Object)
    - memo (dict)
    """
    assert isinstance(new_object_metadata, abc.ObjectMeta)
    object_metadata: Optional[abc.Meta] = meta.read(object_class)
    assert isinstance(object_metadata, (abc.ObjectMeta, NoneType))
    _update_object_meta(object_metadata, new_object_metadata, memo)
    # Recreate the object class
    updated_object_class: type = from_meta(
        object_class.__name__,
        object_metadata,
        module=object_class.__module__,
        docstring=object_class.__doc__,
    )
    # Apply the rest of the changes to the existing model class
    setattr(
        object_class, "__init__", getattr(updated_object_class, "__init__")
    )
    setattr(object_class, "_source", get_source(updated_object_class))


def _update_array_class_from_meta(
    array_class: type,
    new_array_metadata: abc.ArrayMeta,
    memo: Optional[Dict[str, type]] = None,
) -> None:
    """
    This function merges new metadata into an existing array model's metadata.

    Parameters:

    - array_class (type)
    - new_array_metadata (sob.meta.Array)
    - memo (dict)
    """
    assert memo is not None
    assert isinstance(new_array_metadata, abc.ArrayMeta)
    array_metadata: abc.Meta = meta.writable(array_class)
    assert isinstance(array_metadata, abc.ArrayMeta)
    _update_array_meta(array_metadata, new_array_metadata, memo=memo)
    # Since the two array classes have the same name--the class
    # definition/declaration won't change, so updating the metadata is all that
    # is necessary


def _update_dictionary_class_from_meta(
    dictionary_class: type,
    new_dictionary_metadata: abc.DictionaryMeta,
    memo: Optional[Dict[str, type]] = None,
) -> None:
    """
    This function merges new metadata into an existing array model's metadata.

    Parameters:

    - dictionary_class (type)
    - new_array_metadata (sob.meta.Array)
    - memo (dict)
    """
    assert memo is not None
    assert isinstance(new_dictionary_metadata, abc.ArrayMeta)
    dictionary_metadata: abc.Meta = meta.writable(dictionary_class)
    assert isinstance(dictionary_metadata, abc.DictionaryMeta)
    _update_dictionary_meta(
        dictionary_metadata, new_dictionary_metadata, memo=memo
    )
    # Since the two dictionary classes have the same name--the class
    # definition/declaration won't change, so updating the metadata is all that
    # is necessary


def _update_model_class_from_meta(
    model_class: type,
    new_metadata: Union[abc.ArrayMeta, abc.ObjectMeta, abc.DictionaryMeta],
    memo: Optional[Dict[str, type]] = None,
) -> None:
    """
    This function merges new metadata into an existing model's metadata.

    Parameters:

    - model_class (type)
    - new_metadata (sob.meta.Array|sob.meta.Object)
    - memo (dict)
    """
    assert memo is not None
    assert issubclass(model_class, (abc.Array, abc.Object, abc.Dictionary))
    # Update the model metadata in-place
    if issubclass(model_class, abc.Array):
        assert isinstance(new_metadata, abc.ArrayMeta)
        _update_array_class_from_meta(model_class, new_metadata, memo=memo)
    elif issubclass(model_class, abc.Object):
        assert issubclass(model_class, abc.Object)
        assert isinstance(new_metadata, abc.ObjectMeta)
        _update_object_class_from_meta(model_class, new_metadata, memo=memo)
    else:
        assert issubclass(model_class, abc.Dictionary)
        assert isinstance(new_metadata, abc.DictionaryMeta)
        _update_dictionary_class_from_meta(
            model_class, new_metadata, memo=memo
        )


def class_name_from_pointer(pointer: str) -> str:
    """
    This function returns a class name based on the
    [sob.thesaurus.Thesaurus](#Thesaurus) key of the
    [sob.thesaurus.Synonyms](#Synonyms) instance to which an element belongs,
    combined with the *JSON pointer* of the applicable element. This function
    can be substituted for another, when generating a module from a thesaurus,
    by passing a function to the `name` parameter of
    [sob.thesaurus.Thesaurus.get_module_source](#Thesaurus-getmodulesource),
    [sob.thesaurus.Thesaurus.get_module](#Thesaurus-getmodule), or
    [sob.thesaurus.Thesaurus.save_module](#Thesaurus-savemodule).

    Parameters:

    - pointer (str): The synonyms key + JSON pointer of the element for which
      the class is being generated.
    """
    return class_name(
        f"{pointer[:-2]}/item"
        if pointer.endswith("/0")
        else pointer.replace("/0/", "/item/")
    )


def _get_models_from_meta(
    pointer: str,
    metadata: Union[abc.ArrayMeta, abc.ObjectMeta, abc.DictionaryMeta],
    module: str = "__main__",
    memo: Optional[Dict[str, type]] = None,
    name: Callable[[str], str] = class_name_from_pointer,
) -> List[type]:
    """
    This function generates and updates classes from metadata.

    Parameters:

    - pointer (str)
    - metadata (sob.meta.Array|sob.meta.Object)
    - module (str)
    - memo (dict)
    """
    assert memo is not None
    new_models: List[type] = []
    # If the object has no attributes, interpret it an an empty dictionary
    if isinstance(metadata, abc.ObjectMeta) and not metadata.properties:
        metadata = meta.Dictionary()
    # If a model of the same pointer already exists, we update it to
    # reflect our metadata, otherwise--we create a new model
    if pointer in memo:
        _update_model_class_from_meta(memo[pointer], metadata, memo=memo)
    else:
        new_model: type = from_meta(name(pointer), metadata, module=module)
        memo[pointer] = new_model
        new_models.append(new_model)
    return new_models


def _is_base64(value: Any) -> bool:
    """
    Test to see if `value` can be interpreted as base-64 encoded binary data.
    """
    if isinstance(value, str):
        try:
            b64decode(bytes(value, encoding="utf-8"), validate=True)
            return True
        except binascii.Error:
            pass
    return False


@lru_cache(maxsize=128)
def _str_date_or_datetime(value: str) -> type:
    """
    Test to see if `value` can be interpreted as an ISO-8601 encoded `date` or
    `datetime`.
    """
    try:
        timestamp: datetime = parse_date(value)
        if (
            timestamp.hour
            or timestamp.minute
            or timestamp.second
            or timestamp.microsecond
            or "T" in value
            or " " in value
            or "Z" in value
        ):
            return datetime
        else:
            return date
    except ParseError:
        return str


def _is_datetime_str(value: Any) -> bool:
    """
    Test to see if `value` can be interpreted as an ISO-8601 encoded
    `datetime`.

    >>> _is_datetime_str("1999-12-31T00:00:00.000")
    """
    return isinstance(value, str) and _str_date_or_datetime(value) is datetime


def _is_date_str(value: Any) -> bool:
    """
    Test to see if `value` can be interpreted as an ISO-8601 encoded
    `date`.
    """
    return isinstance(value, str) and _str_date_or_datetime(value) is date


def _is_not_null_or_none(item: Any) -> bool:
    return item not in (None, NULL)


@collections.abc.MutableSet.register
class Synonyms:
    """
    This class contains deserialized data, implied to represent variations of
    one type of entity, and is used to infer a model for that entity.
    """

    __slots__ = ("_type", "_nullable", "_set")

    def __init__(
        self, items: Iterable[Union[abc.Readable, abc.MarshallableTypes]] = ()
    ) -> None:
        self._type: Optional[type] = None
        self._nullable: bool = False
        self._set: Set[abc.MarshallableTypes] = set()
        if items:
            self.__ior__(items)

    def add(self, item: Union[abc.Readable, abc.MarshallableTypes]) -> None:
        """
        This method adds a synonymous item to the set. If the item is a
        file-like (input/output) object, that object is first read,
        deserialized, and unmarshalled.

        Parameters:

        - item ({}): A file-like or a JSON-serializable python object.
        """.format(
            "|".join(
                qualified_name(item_type)
                for item_type in ((abc.Readable,) + abc.MARSHALLABLE_TYPES)
            )
        )
        assert isinstance(item, (abc.Readable,) + abc.MARSHALLABLE_TYPES)
        if isinstance(item, abc.Readable):
            # Deserialize and unmarshal file-like objects
            item = unmarshal(detect_format(_read(item))[0])
        elif isinstance(item, Iterable) and not isinstance(
            item, (str, abc.Model, Mapping)
        ):
            if isinstance(item, Iterable) and not isinstance(item, Sequence):
                item = tuple(item)
            # Unmarshal items which appear to not have been part of an
            # unmarshalled container
            item = unmarshal(item)
        if isinstance(item, Null):
            self._nullable = True
        elif item is not None:
            if self._type is None:
                self._type = type(item)
            else:
                # If there is a data type discrepancy between `float` and
                # `int`, use `float`.
                if (
                    issubclass(self._type, (int, float))
                    and isinstance(item, (int, float))
                    and (type(item) is not self._type)
                ):
                    self._type = float
                elif not isinstance(item, self._type):
                    # If there is a discrepancy where a data type can
                    # be either simple of a container, we infer
                    # unrestricted polymorphism, and indicate this by
                    # setting `._type` to `object`
                    self._type = object
            assert isinstance(item, MARSHALLABLE_TYPES)
            self._set.add(item)  # type: ignore

    def discard(self, item: abc.MarshallableTypes) -> None:
        self_set: Set[abc.MarshallableTypes] = copy(self._set)
        self_set.discard(item)
        self.clear()
        self.__ior__(self_set)

    def remove(self, item: abc.MarshallableTypes) -> None:
        self_set: Set[abc.MarshallableTypes] = copy(self._set)
        self_set.remove(item)
        self.clear()
        self.__ior__(self_set)

    def pop(self) -> abc.MarshallableTypes:
        self_set: Set[abc.MarshallableTypes] = copy(self._set)
        item: abc.MarshallableTypes = self_set.pop()
        self.clear()
        self.__ior__(self_set)
        return item

    def clear(self) -> None:
        self._type = None
        self._nullable = False
        self._set.clear()

    def union(
        self, other: Iterable[Union[abc.Readable, abc.MarshallableTypes]]
    ) -> "Synonyms":
        """
        This method returns an instance of `Synonyms` which incorporates
        all (non-redundant) items from both `self` and `other`.
        """
        new_synonyms: Synonyms = copy(self)
        new_synonyms |= other
        return new_synonyms

    def __ior__(
        self, other: Iterable[Union[abc.Readable, abc.MarshallableTypes]]
    ) -> "Synonyms":
        assert_is_instance("other", other, Iterable)
        assert_not_is_instance("other", other, (Mapping, abc.Dictionary))
        item: Union[abc.Readable, abc.MarshallableTypes]
        for item in other:
            self.add(item)
        return self

    def __or__(
        self, other: Iterable[Union[abc.Readable, abc.MarshallableTypes]]
    ) -> "Synonyms":
        return copy(self).__ior__(other)

    @staticmethod
    def _get_set(
        other: Iterable[Union[abc.Readable, abc.MarshallableTypes]]
    ) -> Set[abc.MarshallableTypes]:
        return (
            other._set
            if isinstance(other, Synonyms)
            else other
            if isinstance(other, Set)
            else Synonyms(other)._set
        )

    def __iand__(
        self,
        other: Union[
            "Synonyms", Iterable[Union[abc.Readable, abc.MarshallableTypes]]
        ],
    ) -> "Synonyms":
        other_set: Set[abc.MarshallableTypes] = self._get_set(other)
        self_set: Set[abc.MarshallableTypes] = copy(self._set)
        self.clear()
        return self.__ior__(self_set & other_set)

    def __ixor__(
        self,
        other: Union[
            "Synonyms", Iterable[Union[abc.Readable, abc.MarshallableTypes]]
        ],
    ) -> "Synonyms":
        self_set: Set[abc.MarshallableTypes] = copy(self._set)
        other_set: Set[abc.MarshallableTypes] = self._get_set(other)
        self.clear()
        if self_set is not other_set:
            self.__ior__(self_set ^ other_set)
        return self

    def __isub__(
        self,
        other: Union[
            "Synonyms", Iterable[Union[abc.Readable, abc.MarshallableTypes]]
        ],
    ) -> "Synonyms":
        self_set: Set[abc.MarshallableTypes] = self._set
        other_set: Set[abc.MarshallableTypes] = self._get_set(other)
        self.clear()
        if self_set is not other_set:
            self.__ior__(self_set.__isub__(other_set))
        return self

    def __sub__(
        self, other: Union["Synonyms", Iterable[abc.MarshallableTypes]]
    ) -> "Synonyms":
        other_set: Set[abc.MarshallableTypes] = self._get_set(other)
        return self.__class__(self._set.__sub__(other_set))

    def __xor__(
        self,
        other: Union[
            "Synonyms", Iterable[Union[abc.Readable, abc.MarshallableTypes]]
        ],
    ) -> "Synonyms":
        other_set: Set[abc.MarshallableTypes] = self._get_set(other)
        return self.__class__(self._set.__xor__(other_set))

    def __copy__(self) -> "Synonyms":
        new_synonyms: Synonyms = self.__class__()
        new_synonyms._set = copy(self._set)
        new_synonyms._type = self._type
        new_synonyms._nullable = self._nullable
        return new_synonyms

    def __deepcopy__(self, memo: Optional[dict] = None) -> "Synonyms":
        new_synonyms: Synonyms = self.__class__()
        new_synonyms._set = deepcopy(self._set, memo=memo)
        new_synonyms._type = self._type
        new_synonyms._nullable = self._nullable
        return new_synonyms

    def _get_type(self) -> type:
        assert self._type is not None
        data_type: type = self._type
        # Determine if this is a string encoded to represent a `date`,
        # `datetime`, or base-64 encoded `bytes`.
        if issubclass(self._type, str):
            if all(map(_is_base64, filter(_is_not_null_or_none, self))):
                data_type = bytes
            elif all(
                map(_is_datetime_str, filter(_is_not_null_or_none, self))
            ):
                data_type = datetime
            elif all(map(_is_date_str, filter(_is_not_null_or_none, self))):
                data_type = date
        else:
            # This function should only be invoked for simple types, so we
            # first verify the data type is not a container
            assert not issubclass(self._type, (abc.Model, Mapping, Collection))
        assert data_type is not None
        return data_type

    def _get_property_names_values(
        self,
    ) -> abc.OrderedDict[str, List[abc.MarshallableTypes]]:
        keys_values: abc.OrderedDict[
            str, List[abc.MarshallableTypes]
        ] = collections.OrderedDict()
        item: abc.MarshallableTypes
        for item in self:
            try:
                assert isinstance(item, (Mapping, abc.Dictionary))
            except AssertionError:
                raise IsInstanceAssertionError(
                    "item", item, (Mapping, abc.Dictionary)
                )
            value: abc.MarshallableTypes
            for key, value in item.items():
                if key not in keys_values:
                    keys_values[key] = []
                keys_values[key].append(value)
        return keys_values

    def _get_dictionary_models(
        self,
        pointer: str,
        module: str = "__main__",
        name: Callable[[str], str] = class_name_from_pointer,
        memo: Optional[Dict[str, type]] = None,
    ) -> Iterable[type]:
        metadata: abc.DictionaryMeta = meta.Dictionary()
        metadata.value_types: abc.Types = meta.MutableTypes()  # type: ignore
        key: str
        property_name_: str
        values: List[abc.MarshallableTypes]
        model_class: type
        for model_class in _get_models_from_meta(
            pointer, metadata, module=module, memo=memo, name=name
        ):
            yield model_class

    def _get_object_models(
        self,
        pointer: str,
        module: str = "__main__",
        name: Callable[[str], str] = class_name_from_pointer,
        memo: Optional[Dict[str, type]] = None,
    ) -> Iterable[type]:
        metadata: abc.ObjectMeta = meta.Object()
        metadata.properties = meta.Properties()  # type: ignore
        key: str
        property_name_: str
        values: List[abc.MarshallableTypes]
        for key, values in self._get_property_names_values().items():
            property_name_ = property_name(key)
            item_type: Optional[type] = None
            is_model: bool = False
            property_synonyms: Synonyms = type(self)(values)
            for item_type in property_synonyms._get_types(
                pointer=f"{pointer}/{escape_reference_token(property_name_)}",
                module=module,
                memo=memo,
                name=name,
            ):
                if issubclass(item_type, abc.Model):
                    is_model = True
                    yield item_type
                else:
                    is_model = False
            if item_type:
                if is_model:
                    metadata.properties[property_name_] = Property(
                        name=key,
                        types=[item_type]
                        + ([Null] if property_synonyms._nullable else []),
                    )
                elif property_synonyms._nullable:
                    metadata.properties[property_name_] = Property(
                        name=key, types=[TYPES_PROPERTIES[item_type](), Null]
                    )
                else:
                    metadata.properties[property_name_] = TYPES_PROPERTIES[
                        item_type
                    ](name=key)
            else:
                metadata.properties[property_name_] = Property(name=key)
        model_class: type
        for model_class in _get_models_from_meta(
            pointer, metadata, module=module, memo=memo, name=name
        ):
            yield model_class

    def _get_array_models(
        self,
        pointer: str,
        module: str = "__main__",
        name: Callable[[str], str] = class_name_from_pointer,
        memo: Optional[Dict[str, type]] = None,
    ) -> Iterable[type]:
        unified_items: Synonyms = type(self)()
        items: abc.MarshallableTypes
        for items in self:
            assert isinstance(items, Iterable)
            unified_items |= items
        item_type: Optional[type] = None
        for item_type in unified_items._get_types(
            pointer=f"{pointer}/0", module=module, memo=memo, name=name
        ):
            yield item_type
        metadata: abc.ArrayMeta = meta.Array(
            item_types=(None if item_type is None else [item_type])
        )
        array_type: type
        for array_type in _get_models_from_meta(
            pointer, metadata, module=module, memo=memo, name=name
        ):
            yield array_type

    def _get_types(
        self,
        pointer: str,
        module: str = "__main__",
        name: Callable[[str], str] = class_name_from_pointer,
        memo: Optional[Dict[str, type]] = None,
    ) -> Iterable[type]:
        # `_memo` holds a dictionary of all classes which have been created,
        # and is passed recursively to facilitate de-duplication
        memo_is_new: bool = False
        if memo is None:
            memo = {}
            memo_is_new = True
        type_iterator: Iterable[type]
        if self._type is None:
            type_iterator = []
        elif issubclass(self._type, (Mapping, abc.Object, abc.Dictionary)):
            type_iterator = self._get_object_models(
                pointer, module=module, memo=memo, name=name
            )
        elif issubclass(self._type, (Iterable, abc.Array)) and not issubclass(
            self._type, str
        ):
            type_iterator = self._get_array_models(
                pointer, module=module, memo=memo, name=name
            )
        else:
            data_type: type = self._get_type()
            type_iterator = [] if data_type is object else [data_type]
        # If this was the call which initialized our `_memo`, we want to
        # force the iterator to run, in order to fully update all models before
        # returning them to the user (as some will be updated over the course
        # of traversal when analogous elements are encountered).
        if memo_is_new:
            type_iterator = list(type_iterator)
        type_: type
        for type_ in type_iterator:
            yield type_

    def get_models(
        self,
        pointer: str,
        module: str = "__main__",
        name: Callable[[str], str] = class_name_from_pointer,
    ) -> Iterable[type]:
        """
        Retrieve a sequence of class definitions representing a data model
        capable of describing these synonyms.

        Parameters:

        - pointer (str): A JSON pointer for the top-level model class, used to
          infer class names.
        - module (str): The name of the module in which model classes will be
          defined. This defaults to "__main__".
        - name (str) = sob.thesaurus.class_name_from_pointer:
          A function which accepts one `str` argument—a synonym key
          concatenated with "#" and JSON pointer (for example:
          "key#/body/items/0") and which returns a `str` which will be the
          resulting class name (for example: "KeyBodyItemsItem").
        """
        assert callable(name)
        # This assertion ensures `self` contains data which can be described by
        # a model class.
        assert self._type and not issubclass(
            self._type, (str, bytes, bytearray)
        )
        if self._type is not object:
            assert_is_subclass(
                f"{qualified_name(type(self))}._type",
                self._type,
                (Mapping, abc.Dictionary, Iterable),
            )
        for model_class in self._get_types(
            pointer="{}#".format(quote_plus(pointer, safe="/+")),
            module=module,
            name=name,
        ):
            assert_is_subclass("model_class", model_class, abc.Model)
            yield model_class

    def __len__(self) -> int:
        return self._set.__len__()

    def __iter__(self) -> Iterator[abc.MarshallableTypes]:
        return self._set.__iter__()

    def __contains__(self, item: abc.MarshallableTypes) -> bool:
        return self._set.__contains__(item)

    def __le__(
        self, other: Iterable[Union[abc.Readable, abc.MarshallableTypes]]
    ) -> bool:
        return self._set.__le__(self._get_set(other))

    def __lt__(
        self, other: Iterable[Union[abc.Readable, abc.MarshallableTypes]]
    ) -> bool:
        return self._set.__lt__(self._get_set(other))

    def __gt__(
        self, other: Iterable[Union[abc.Readable, abc.MarshallableTypes]]
    ) -> bool:
        return self._set.__gt__(self._get_set(other))

    def __ge__(
        self, other: Iterable[Union[abc.Readable, abc.MarshallableTypes]]
    ) -> bool:
        return self._set.__ge__(self._get_set(other))

    def __eq__(self, other: Any) -> bool:
        return type(other) is type(self) and self._set.__eq__(
            self._get_set(other)
        )

    def __and__(
        self, other: Iterable[Union[abc.Readable, abc.MarshallableTypes]]
    ) -> "Synonyms":
        return copy(self).__iand__(other)

    def isdisjoint(
        self, other: Iterable[Union[abc.Readable, abc.MarshallableTypes]]
    ) -> bool:
        return self._set.isdisjoint(self._get_set(other))


def get_class_meta_attribute_assignment_source(
    class_name_: str,
    attribute_name: str,
    metadata: abc.Meta,
) -> str:
    """
    This function generates source code for setting a metadata attribute on
    a class.

    Parameters:

    - class_name (str): The name of the class to which we want to assign a
      metadata attribute.
    - attribute_name (str): The name of the attribute we want to assign.
    - metadata (sob.abc.Meta): The metadata from which to take the assigned
      value.
    """
    writable_function_name: str = "{}.meta.{}_writable".format(
        _parent_module_name,
        (
            "object"
            if isinstance(metadata, abc.ObjectMeta)
            else "array"
            if isinstance(metadata, abc.ArrayMeta)
            else "dictionary"
        ),
    )
    # We insert "  # type: ignore" at the end of the first line where the value
    # is assigned due to mypy issues with properties having getters and setters
    return suffix_long_lines(
        (
            f"{writable_function_name}(  # type: ignore\n"
            f"    {suffix_long_lines(class_name_, -4)}\n"
            f").{attribute_name} = {repr(getattr(metadata, attribute_name))}"
        ),
        -4,
    )


@collections.abc.MutableMapping.register
class Thesaurus:
    """
    TODO
    """

    __slots__ = ("_dict",)

    def __init__(
        self,
        _items: Union[
            Mapping[
                str,
                Union[
                    Iterable[Union[abc.Readable, abc.MarshallableTypes]],
                    Synonyms,
                ],
            ],
            Iterable[
                Tuple[
                    str,
                    Union[
                        Iterable[Union[abc.Readable, abc.MarshallableTypes]],
                        Synonyms,
                    ],
                ]
            ],
            "Thesaurus",
            None,
        ] = None,
        **kwargs: Iterable[Union[abc.Readable, abc.MarshallableTypes]],
    ) -> None:
        self._dict: abc.OrderedDict[str, Synonyms] = collections.OrderedDict()
        key: str
        value: Iterable[Union[abc.Readable, abc.MarshallableTypes]]
        for key, value in collections.OrderedDict(
            *((_items,) if _items else ()), **kwargs
        ).items():
            self[key] = value

    def __setitem__(
        self,
        key: str,
        value: Iterable[Union[abc.Readable, abc.MarshallableTypes]],
    ) -> None:
        if not isinstance(value, Synonyms):
            value = Synonyms(value)
        return self._dict.__setitem__(key, value)

    def __delitem__(self, key: str) -> None:
        self._dict.__delitem__(key)

    def pop(
        self, key: str, default: Union[Synonyms, Undefined] = UNDEFINED
    ) -> Synonyms:
        return self._dict.pop(
            key,
            **(
                {} if isinstance(default, Undefined) else dict(default=default)
            ),
        )

    def popitem(self) -> Tuple[str, Synonyms]:
        return self._dict.popitem()

    def clear(self) -> None:
        self._dict.clear()

    def update(self, **kwargs: Synonyms) -> None:
        key: str
        value: Iterable[Union[abc.Readable, abc.MarshallableTypes]]
        for key, value in kwargs.items():
            self[key] = value

    def setdefault(
        self,
        key: str,
        default: Iterable[Union[abc.Readable, abc.MarshallableTypes]],
    ) -> Synonyms:
        if not isinstance(default, Synonyms):
            default = Synonyms(default)
        return self._dict.setdefault(key, default)

    def __getitem__(self, key: str) -> Synonyms:
        try:
            return self._dict.__getitem__(key)
        except KeyError:
            self[key] = Synonyms()
            return self[key]

    def get(
        self, key: str, default: Union[Undefined, Synonyms] = UNDEFINED
    ) -> Synonyms:
        return self._dict.get(
            key,
            **(
                {} if isinstance(default, Undefined) else dict(default=default)
            ),
        )

    def __contains__(self, key: str) -> bool:
        return self._dict.__contains__(key)

    def keys(self) -> KeysView[str]:
        return self._dict.keys()

    def items(self) -> ItemsView[str, Synonyms]:
        return self._dict.items()

    def values(self) -> ValuesView[Synonyms]:
        return self._dict.values()

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, self.__class__):
            return self._dict == other._dict
        return False

    def __copy__(self) -> "Thesaurus":
        return self.__class__(copy(self._dict))

    def __reversed__(self) -> Iterator[str]:
        return reversed(self._dict)

    def __deepcopy__(self, memo: Optional[dict] = None) -> "Thesaurus":
        return self.__class__(deepcopy(self._dict, memo=memo))

    def __iadd__(self, other: "Thesaurus") -> "Thesaurus":
        assert isinstance(other, Thesaurus)
        key: str
        value: Synonyms
        for key, value in other.items():
            self[key] |= value
        return self

    def __add__(self, other: "Thesaurus") -> "Thesaurus":
        return copy(self).__iadd__(other)

    def get_models(
        self,
        module: str = "__main__",
        name: Callable[[str], str] = class_name_from_pointer,
    ) -> Iterable[type]:
        key: str
        synonyms: Synonyms
        for key, synonyms in self.items():
            model_class: type
            for model_class in synonyms.get_models(
                key, module=module, name=name
            ):
                yield model_class

    def _get_module_source(
        self,
        module_name: str = "__main__",
        name: Callable[[str], str] = class_name_from_pointer,
    ) -> str:
        class_names_metadata: abc.OrderedDict[
            str, Union[abc.ObjectMeta, abc.ArrayMeta, abc.DictionaryMeta]
        ] = collections.OrderedDict()
        imports: List[str] = []
        classes: List[str] = []
        metadatas: List[str] = []
        model_class: type
        for model_class in self.get_models(module=module_name, name=name):
            imports_str: str
            import_line: str
            source: str = get_source(model_class)
            if "\n\n\n" not in source:
                raise ValueError(source)
            imports_str, source = source.split("\n\n\n")
            for import_line in imports_str.split("\n"):
                if import_line not in imports:
                    imports.append(import_line)
            classes.append(source)
            meta_instance: Optional[abc.Meta] = meta.read(model_class)
            assert isinstance(
                meta_instance,
                (abc.ObjectMeta, abc.ArrayMeta, abc.DictionaryMeta),
            )
            class_names_metadata[model_class.__name__] = meta_instance
        for negative_index, class_name_metadata in enumerate(
            class_names_metadata.items(), -(len(class_names_metadata) - 1)
        ):
            separator: str
            class_name_: str
            metadata: Union[abc.ObjectMeta, abc.ArrayMeta, abc.DictionaryMeta]
            class_name_, metadata = class_name_metadata
            assert isinstance(
                metadata, (abc.ObjectMeta, abc.ArrayMeta, abc.DictionaryMeta)
            )
            if isinstance(metadata, abc.ObjectMeta):
                metadatas.append(
                    get_class_meta_attribute_assignment_source(
                        class_name_, "properties", metadata
                    )
                )
            elif isinstance(metadata, abc.ArrayMeta):
                metadatas.append(
                    get_class_meta_attribute_assignment_source(
                        class_name_, "item_types", metadata
                    )
                )
            else:
                metadatas.append(
                    get_class_meta_attribute_assignment_source(
                        class_name_, "value_types", metadata
                    )
                )
        return "\n".join(
            sorted(
                imports,
                key=lambda line: (
                    1 if line == f"import {_parent_module_name}" else 0
                ),
            )
            + ["\n"]
            + classes
            + metadatas
        )

    def get_module_source(
        self, name: Callable[[str], str] = class_name_from_pointer
    ) -> str:
        """
        This method generates and returns the source code for a module
        defining data models applicable to the data contained in this
        thesaurus.

        Parameters:

        - name (str) = sob.thesaurus.class_name_from_pointer:
          A function which accepts one `str` argument—a synonym key
          concatenated with "#" and JSON pointer (for example:
          "key#/body/items/0") and which returns a `str` which will be the
          resulting class name (for example: "KeyBodyItemsItem").
        """
        return self._get_module_source("__main__", name=name)

    def get_module(
        self, name: Callable[[str], str] = class_name_from_pointer
    ) -> ModuleType:
        """
        This method generates and returns a module defining data models
        applicable to the data contained in this thesaurus. This module is not
        suitable for writing out for static use--use `Thesaurus.save_module`
        to generate and write a model suitable for static use.

        Parameters:

        - name (str) = sob.thesaurus.class_name_from_pointer:
          A function which accepts one `str` argument—a synonym key
          concatenated with "#" and JSON pointer (for example:
          "key#/body/items/0") and which returns a `str` which will be the
          resulting class name (for example: "KeyBodyItemsItem").
        """
        # For pickling to work, the `__module__` variable needs to be set to
        # the calling module.
        module_name: str = calling_module_name(2)
        module: ModuleType = ModuleType(module_name)
        exec(self._get_module_source(module_name, name=name), module.__dict__)
        return module

    def save_module(
        self,
        path: str,
        name: Callable[[str], str] = class_name_from_pointer,
    ) -> None:
        """
        This method generates and saves the source code for a module
        defining data models applicable to the data contained in this
        thesaurus.

        Parameters:

        - path (str): The file path where the data will be written.
        - name (str) = sob.thesaurus.class_name_from_pointer:
          A function which accepts one `str` argument—a synonym key
          concatenated with "#" and JSON pointer (for example:
          "key#/body/items/0") and which returns a `str` which will be the
          resulting class name (for example: "KeyBodyItemsItem").
        """
        os.makedirs(os.path.dirname(path), exist_ok=True)
        module_source: str = self.get_module_source(name=name)
        with open(path, "w") as module_io:
            module_io.write(f"{module_source}\n")



import sys
from typing import Any, Set

import setuptools  # type: ignore

_INSTALL_REQUIRES: str = "install_requires"


def setup(**kwargs: Any) -> None:
    """
    This `setup` script intercepts arguments to be passed to
    `setuptools.setup` in order to dynamically alter setup requirements
    while retaining a function call which can be easily parsed and altered
    by `setuptools-setup-versions`.
    """
    # Require the package "dataclasses" for python 3.6, but not later
    # python versions (since it's part of the standard library after 3.6)
    if sys.version_info[:2] == (3, 6):
        if _INSTALL_REQUIRES not in kwargs:
            kwargs[_INSTALL_REQUIRES] = []
        kwargs[_INSTALL_REQUIRES].append("dataclasses")
    # Add an "all" extra which includes all extra requirements
    if "extras_require" in kwargs and "all" not in kwargs["extras_require"]:
        all_extras_require: Set[str] = set()
        kwargs["extras_require"]["all"] = []
        for extra_name, requirements in kwargs["extras_require"].items():
            if extra_name != "all":
                for requirement in requirements:
                    if requirement not in all_extras_require:
                        all_extras_require.add(requirement)
                        kwargs["extras_require"]["all"].append(requirement)
    # Pass the modified keyword arguments to `setuptools.setup`
    setuptools.setup(**kwargs)


setup(
    name="sob",
    version="1.18.0",
    description=(
        "A framework for serializing/deserializing JSON/YAML into python "
        "class instances and vice versa"
    ),
    url="https://github.com/davebelais/sob",
    author="David Belais",
    author_email="david@belais.me",
    license="MIT",
    include_package_data=True,
    zip_safe=False,
    python_requires="~=3.6",
    keywords="rest api serialization serialize",
    packages=["sob", "sob.utilities"],
    package_data={"sob": ["py.typed"], "sob.utilities": ["py.typed"]},
    install_requires=["pyyaml>=3.10", "iso8601~=0.1", "more-itertools~=8.10"],
    setup_requires=["setuptools"],
    extras_require={
        "test": [
            "black~=21.9b0",
            "wheel~=0.36.2",
            "mypy~=0.910",
            "tox~=3.20",
            "pytest~=5.4",
            "flake8~=3.9",
            "types-PyYAML~=5.4",
            "setuptools-setup-versions>=1.18.0,<2",
            "readme-md-docstrings>=0.1.0,<1",
            "twine~=3.2",
            "daves-dev-tools~=0.9",
        ],
        "dev": ["readme-md-docstrings~=0.1.0,<2"],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
    ],
)
